#!/usr/bin/env python
"""GenomeArrays map quantitative data or :term:`read alignments` to genomic positions.

.. contents::
   :local:

Summary
-------

GenomeArrays are typically accessed using |SegmentChains|.

The :meth:`~plastid.genomics.roitools.SegmentChain.get_counts` method of
|SegmentChain| and |Transcript| fetches arrays of data from GenomeArrays. The
result is a :class:`numpy array <numpy.ndarray>` in which each position
corresponds to a position in the |SegmentChain|, moving from 5' to 3', in the
spliced |SegmentChain| or |Transcript|. For a |Transcript| called
`my_transcript`::

    >>> # open a BAM file, map reads 15 nucleotides from 3' end
    >>> genome_array = BAMGenomeArray("some_file.bam")
    >>> genome_array.set_mapping(ThreePrimeMapFactory(offset=15))

    >>> my_data = my_transcript.get_counts(genome_array)
    >>> my_data
        array([ 95, 133,  87, 491, 100,  73, 308, 217, 447, 204, 245, 264, 395,
                76, 387, 319,  43, 171, 101, 152,  58, 167,  53, 397, 228,  44,
               402, 285, 480,  55,   9,  51, 358, 307, 144, 182, 293, 455, 403,
               375, 443, 442, 196, 133, 238, 268,  70,  91, 172, 490, 346, 294,
               117, 424, 180,  14, 400, 414, 257, 127, 344, 307, 435, 152, 458,
               423,  60, 173, 362, 277, 393, 297, 377, 357,  68, 143, 467,   7,
                86, 450, 230,  62, 278,  65, 346, 103, 185,   7, 350, 162, 461,
               274, 205, 185, 209, 275, 190, 434,  10, 241])


Alternatively, |SegmentChains| and |GenomicSegments| can be used directly as
dictionary keys::

    >>> my_data = genome_array[my_transcript]
    >>> my_data
        array([ 95, 133,  87, 491, 100,  73, 308, 217, 447, 204, 245, 264, 395,
                76, 387, 319,  43, 171, 101, 152,  58, 167,  53, 397, 228,  44,
               402, 285, 480,  55,   9,  51, 358, 307, 144, 182, 293, 455, 403,
               375, 443, 442, 196, 133, 238, 268,  70,  91, 172, 490, 346, 294,
               117, 424, 180,  14, 400, 414, 257, 127, 344, 307, 435, 152, 458,
               423,  60, 173, 362, 277, 393, 297, 377, 357,  68, 143, 467,   7,
                86, 450, 230,  62, 278,  65, 346, 103, 185,   7, 350, 162, 461,
               274, 205, 185, 209, 275, 190, 434,  10, 241])


Module contents
---------------

Several implementations are provided. These are tailored to data stored in 
specific formats, or for specific purposes. For example, mutable GenomeArrays
allow values to be changed in-place, e.g. via mathematical operations or manual
setting.


====================   ==================================   =====================   ===================================================
**Implementation**     **Format of data**                   **Mutable**             **Use of mapping functions**
--------------------   ----------------------------------   ---------------------   ---------------------------------------------------
|BAMGenomeArray|       `BAM`_ file(s)                       No                      Inline; may be changed without reloading data

|BigWigGenomeArray|    `BigWig`_ file(s)                    No                      n/a

|GenomeArray|          `bowtie`_, `wiggle`_, `bedGraph`_,   Yes                     On import from `bowtie`_ files only. Data must be
                       or data in memory                                            reimported to change mapping

|SparseGenomeArray|    `bowtie`_, `wiggle`_, `bedGraph`_,   Yes                     On import from `bowtie`_ files only. Data must be
                       or data in memory                                            reimported to change mapping
====================   ==================================   =====================   ===================================================


Extended summary & examples
---------------------------


Mapping functions
.................

GenomeArrays use :term:`Mapping functions <mapping rule>` to extract the biology
of interest from read alignments, and to vectorize these data over genomic
positions.

Depending on the data type and the purpose of the experiment, different functions
may be appropriate. The following mapping rules are provided, although users are
encouraged to provide their own if needed. These include:


================================   ===================================================    =====================================
**Mapping strategy**               **Description**                                        **Sample use**
--------------------------------   ---------------------------------------------------    -------------------------------------

Fiveprime                          Each read alignment is mapped to its 5' end, or        RNA-seq, ribosome profiling, DMS-seq
                                   at a fixed distance from its 5' end. 

Variable fiveprime                 Each read alignment is mapped at a fixed distance      Ribosome profiling
                                   from its 5' end, where the distance is determined
                                   by the length of the alignment

Stratified variable fiveprime      Like variable fiveprime mapping, except reads are      Ribosome profiling
                                   stratified by length into separate rows of an array

Threeprime                         Each read alignment is mapped to its 5' end, or        RNA-seq, ribosome profiling
                                   at a fixed distance from its 5' end. 

Center/entire                      Zero or more positions are trimmed from each end       Ribosome profiling with micrococcal
                                   of the read alignment, and the remaining `N`           nuclease, RNA-seq
                                   positions are incremented by `1/N`
================================   ===================================================    =====================================


Mapping functions are rarely called directly. Instead, they are invoked via
:meth:`BAMGenomeArray.set_mapping() <plastid.genomics.genome_array.BAMGenomeArray.set_mapping>`:

 .. code-block:: python

    # open BAM file
    >>> ga = BAMGenomeArray("some_file.bam")

    # set mapping 15 bases from 3' end of read
    >>> ga.set_mapping(ThreePrimeMapFactory(offset=15))

    # set mapping to 5' end of read
    >>> ga.set_mapping(FivePrimeMapFactory())

    # cut 3 nt off each end of the read, and map fractionally over the rest
    >>> ga.set_mapping(CenterMapFactory(nibble=12))

    # access reads in some region
    >>> ga[GenomicSegment("chrI",10000,100000,"+")]
    # output is numpy array

For further details see :mod:`~plastid.genomics.map_factories` and
:doc:`/concepts/mapping_rules`


Export to browser tracks
........................

GenomeArrays can export data to `wiggle`_ or `bedGraph`_ formats. To export 
forward-strand data to `some_file.wig`::

    >>> genome_array.to_bedgraph(open("some_file.wig","w"),"my_track","+")


Normalization
.............
        
GenomeArrays also offer utilities for summing and normalization::

    >>> genome_array.sum() # get sum
    71292045
    
    # set sum to an arbitrary value- can be useful for normalizing between
    # multiple alignment files
    >>> genome_array.set_sum(61241234324)
    >>> genome_array.sum()
    61241234324
    
    # reset sum to data in file
    >>> genome_array.reset_sum()
    >>> genome_array.sum()
    71292045
    
    # with set_normalize(), all arrays will be normalized to counts per million
    # relative to the current sum
    >>> genome_array.set_normalize(True)
    >>> my_transcript.get_counts(genome_array)
    array([ 1.33254699,  1.86556579,  1.22033251,  6.88716392,  1.40268104,
            1.02395716,  4.32025761,  3.04381786,  6.26998426,  2.86146933,
            3.43656855,  3.70307795,  5.54059012,  1.06603759,  5.42837563,
            4.47455253,  0.60315285,  2.39858458,  1.41670785,  2.13207518,
            0.813555  ,  2.34247734,  0.74342095,  5.56864374,  3.19811278,
            0.61717966,  5.63877779,  3.99764097,  6.732869  ,  0.77147457,
            0.12624129,  0.71536733,  5.02159813,  4.3062308 ,  2.0198607 ,
            2.5528795 ,  4.10985545,  6.38219874,  5.6528046 ,  5.26005391,
            6.21387702,  6.19985021,  2.74925484,  1.86556579,  3.33838088,
            3.75918519,  0.98187673,  1.27643975,  2.41261139,  6.87313711,
            4.85327641,  4.12388226,  1.64113682,  5.94736762,  2.52482588,
            0.19637535,  5.61072417,  5.80709952,  3.60489028,  1.78140492,
            4.82522279,  4.3062308 ,  6.10166253,  2.13207518,  6.42427917,
            5.93334081,  0.84160863,  2.4266382 ,  5.07770537,  3.88542649,
            5.5125365 ,  4.1659627 ,  5.28810753,  5.00757132,  0.95382311,
            2.00583389,  6.55052047,  0.09818767,  1.2063057 ,  6.31206469,
            3.2261664 ,  0.86966225,  3.8994533 ,  0.91174268,  4.85327641,
            1.44476147,  2.59495993,  0.09818767,  4.90938365,  2.27234329,
            6.46635961,  3.84334606,  2.87549614,  2.59495993,  2.93160338,
            3.85737287,  2.66509398,  6.08763572,  0.1402681 ,  3.38046131])

    # similarly, with set_normalize() on, browser tracks we'll be normalized 
    >>> genome_array.to_bedgraph(open("some_file_normalized.wig","w"),"my_normalized_track","+")

"""
__date__ = "May 3, 2011"
__author__ = "joshua"
from abc import abstractmethod
import itertools
import operator
import copy
import numpy
import scipy.sparse
import pysam

from collections import OrderedDict

from plastid.readers.wiggle import WiggleReader
from plastid.readers.bowtie import BowtieReader
from plastid.readers.bigwig import BigWigReader
from plastid.genomics.roitools import GenomicSegment, SegmentChain
from plastid.util.services.mini2to3 import xrange, ifilter
from plastid.util.services.exceptions import DataWarning, warn
from plastid.util.io.openers import NullWriter, multiopen

from plastid.genomics.map_factories import *

MIN_CHR_SIZE = int(10 * 1e6)  # 10 Mb minimum size for unspecified chromosomes

#===============================================================================
# INDEX: Mapping functions for GenomeArray and SparseGenomeArray.
#        these are used by add_from_bowtie() and add_from_tagalign()
#        to map read alignments to specific sites.
#
#        See function documentation for more details
#===============================================================================


def center_map(feature, **kwargs):
    """Center-mapping function used as an argument to :py:meth:`GenomeArray.add_from_bowtie`.
    A user-specified number of bases is optionally removed from each side of each
    read alignment, and the `N` remaining bases are each apportioned `1/N`
    of the read count, so that the entire read is counted once.
    
    Parameters
    ----------
    feature : |SegmentChain|
        Ungapped genomic alignment
        
    kwargs['value'] : float, optional
        Total value to divide over aligning positions (Default: `1.0`)
        
    kwargs['nibble'] : int, optional
        Positions to remove from each end before mapping (Default: `0`)
        
    kwargs['offset'] : int, optional
        Mapping offset, if any, from 5' end of read (Default: `0`)
    
    Returns
    -------
    list
        tuples of `(GenomicSegment,float value over segment)`
    """
    nibble = kwargs["nibble"]
    read_length = feature.length
    if read_length <= 2 * nibble:
        warn(
            "File contains read alignments shorter (%s nt) than `2*'nibble'` value of %s nt. Ignoring these."
            % (read_length, 2 * nibble), DataWarning
        )
        return []

    # yapf: disable
    span = feature.spanning_segment
    strand = span.strand
    chrom  = span.chrom
    offset = kwargs.get("offset", 0)
    value  = float(kwargs.get("value", 1.0))
    sign   = -1 if strand == "-" else 1
    start  = span.start + nibble + sign * offset
    end    = span.end - nibble + sign * offset
    seg    = GenomicSegment(chrom, start, end, strand)
    frac   = value / len(seg)
    return [(seg, frac)]


def five_prime_map(feature, **kwargs):
    """Fiveprime mapping function used as an argument to :py:meth:`GenomeArray.add_from_bowtie`.
    Reads are mapped at a user-specified offset from the fiveprime end of the alignment.
    
    Parameters
    ----------
    feature : |SegmentChain|
        Ungapped genomic alignment
        
    kwargs['value'] : float or int, optional
        Value to apportion (Default: `1`)
        
    kwargs['offset'] : int, optional
        Mapping offset, if any, from 5' toward 3' end of read
    
    Returns
    -------
    list
        tuples of `(GenomicSegment,float value over segment)`
    """
    offset = kwargs.get("offset", 0)
    read_length = feature.length
    if offset > read_length:
        warn(
            "File contains read alignments shorter (%s nt) than offset (%s nt). Ignoring." %
            (read_length, offset), DataWarning
        )
        return []

    value = kwargs.get("value", 1.0)
    span = feature.spanning_segment
    strand = span.strand
    if strand in ("+", "."):
        start = span.start + offset
    else:
        start = span.end - 1 - offset
    seg = GenomicSegment(span.chrom, start, start + 1, strand)
    return [(seg, value)]


def three_prime_map(feature, **kwargs):
    """Threeprime mapping function used as an argument to :py:meth:`GenomeArray.add_from_bowtie`.
    Reads are mapped at a user-specified offset from the threeprime end of the alignment.
    
    Parameters
    ----------
    feature : |SegmentChain|
        Ungapped genomic alignment
        
    kwargs['value'] : float or int, optional
        Value to apportion (Default: `1`)
        
    kwargs['offset'] : int, optional
        Mapping offset, if any, from 3' toward 5' end of read.
    
    Returns
    -------
    list
        tuples of `(GenomicSegment,float value over segment)`
    """
    offset = kwargs.get("offset", 0)
    read_length = feature.length
    if offset > read_length:
        warn(
            "File contains read alignments shorter (%s nt) than offset (%s nt). Ignoring." %
            (read_length, offset), DataWarning
        )
        return []

    value = kwargs.get("value", 1.0)
    span = feature.spanning_segment
    strand = span.strand
    if strand in ("+", "."):
        start = span.end - 1 - offset
    else:
        start = span.start + offset
    seg = GenomicSegment(span.chrom, start, start + 1, strand)
    return [(seg, value)]


def variable_five_prime_map(feature, **kwargs):
    """Fiveprime variable mapping function used as an argument to :py:meth:`GenomeArray.add_from_bowtie`.
    Reads are mapped at a user-specified offset from the fiveprime end of the alignment.
    The offset for a read of a given length is supplied in `kwargs[offset][readlen]`
    
    Parameters
    ----------
    feature : |SegmentChain|
        Ungapped genomic alignment
        
    kwargs['offset'] : dict, required
        Dictionary mapping read lengths to offsets

    kwargs['value'] : float or int, optional
        Value to apportion (Default: `1`)
    
    Returns
    -------
    list
        tuples of `(GenomicSegment,float value over segment)`
    """
    span = feature.spanning_segment
    strand = span.strand
    value = kwargs.get("value", 1.0)
    offset = kwargs["offset"].get(len(span), kwargs["offset"].get("default", None))
    feature_length = feature.length
    if offset is None:
        warn("No offset for reads of length %s. Ignoring." % feature_length, DataWarning)
        return []
    if offset >= feature_length:
        warn("Offset (%s nt) longer than read length %s. Ignoring" % (offset, feature_length))

    if strand in ("+", "."):
        start = span.start + offset
    else:
        start = span.end - 1 - offset
    seg = GenomicSegment(span.chrom, start, start + 1, strand)
    return [(seg, value)]


#===============================================================================
# GenomeArray classes
#===============================================================================

_DEFAULT_STRANDS = ("+", "-")


class AbstractGenomeArray(object):
    """Abstract base class for all |GenomeArray|-like objects"""

    def __init__(self, chr_lengths=None, strands=None):
        """
        Parameters
        ----------
        chr_lengths : dict
            Dictionary mapping chromosome names to lengths. Suppyling this in
            advance yields considerable speed improvements, as memory can be
            pre-allocated for each chromosomal position. If not provided, a
            minimum chromosome size will be guessed, and chromosomes re-sized
            as needed.
        
        strands : sequence, optional
            Strands for the |AbstractGenomeArray|. (Default: `("+","-")`)
        """
        self._chroms = {}
        self._strands = _DEFAULT_STRANDS if strands is None else strands
        self._sum = None
        self._normalize = False

    def __str__(self):
        return repr(self)

    def __repr__(self):
        stmp = "<%s len=%s sum=%s chroms=" % (self.__class__.__name__, len(self), self.sum())
        stmp += ",".join(self.chroms())
        stmp += " strands=%s" % ",".join(self.strands())
        stmp += ">"
        return stmp

    def __contains__(self, chrom):
        """Return `True` if `chrom` is defined in the array, otherwise False
        
        Returns
        -------
        bool
        """
        return chrom in self._chroms

    def __len__(self):
        """Return size of GenomeArray in nucleotide positions (not base pairs).
        To obtain size in base pairs, divide length by two.
        
        Returns
        -------
        int
        """
        return len(self._strands) * sum(self.lengths().values())

    @abstractmethod
    def __getitem__(self, roi):
        """Retrieve array of counts from a region of interest. The values in
        the returned array are in 5' to 3' with respect to `seg` rather than
        the genome (i.e. are reversed for reverse-strand features).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome
        
        Returns
        -------
        :class:`numpy.ndarray`
            vector of numbers, each position corresponding to a position
            in `seg`, from 5' to 3' relative to `seg`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        pass

    @abstractmethod
    def reset_sum(self):
        """Reset sum to total mapped reads in the GenomeArray"""

    def set_sum(self, val):
        """Set sum used for normalization to an arbitrary value (e.g. from another dataset)
        
        Parameters
        ----------
        val : int or float
            a number
        """
        self._sum = val

    def sum(self):
        """Return the total number of aligned reads or the sum of the 
        quantitative data across all positions in the GenomeArray
        
        Notes
        -----
        The true (i.e. unnormalized) sum is always reported, even if 
        :meth:`set_normalize` is set to True
        
        Returns
        -------
        int or float
        """
        if self._sum is None:
            self.reset_sum()

        return self._sum

    def set_normalize(self, value=True):
        """Toggle normalization of reported values to reads per million mapped in the dataset
        
        Parameters
        ----------
        value : bool
            If `True`, all values fetched will be normalized to reads
            per million. If `False`, all values will not be normalized.
        """
        assert value in (True, False)
        self._normalize = value

    def chroms(self):
        """Return a list of chromosomes in the GenomeArray
        
        Returns
        -------
        list
            Chromosome names as strings
        """
        return self._chroms.keys()

    def strands(self):
        """Return a tuple of strands in the GenomeArray
        
        Returns
        -------
        tuple
            Chromosome strands as strings
        """
        return self._strands

    def lengths(self):
        """Return a dictionary mapping chromosome names to lengths.
        When two strands report different lengths for a chromosome, the
        max length is taken.
        
        Returns
        -------
        dict
            Dictionary mapping chromosome names to chromosome lengths
        """
        d_out = {}.fromkeys(self.keys())
        for key in d_out:
            d_out[key] = max([len(self._chroms[key][X]) for X in self.strands()])
        return d_out


class MutableAbstractGenomeArray(AbstractGenomeArray):
    """Abstract base class for |GenomeArray|-like objects whose values can be
    changed
    """

    @abstractmethod
    def __setitem__(self, seg, val):
        """Set values in |MutableAbstractGenomeArray| over a region of interest.
        
        Parameters
        ----------
        seg : |GenomicSegment| or |SegmentChain|
            Region of interest

        val : int, float, or :class:`numpy.ndarray`
            Scalar or vector of values to set in array over `seg`.
            If a vector, values should be ordered 5'-to-3' relative
            to `seg` rather than (i.e. for a reverse-strand feature,
            position 0 in the vector would correspond to seg.end)
            in the genome
        """
        pass


class BAMGenomeArray(AbstractGenomeArray):
    """
    BAMGenomeArray(*bamfiles,mapping=CenterMapFactory())
    
    A GenomeArray for :term:`read alignments` in `BAM`_ files.
    
    When a user requests :term:`read alignments` or :term:`counts` at
    a set of genomic positions, |BAMGenomeArray| uses a :term:`mapping rule`
    to determine which reads from the `BAM`_ file should be returned and how
    these reads should be mapped to positions in a vector. Reads
    are not mapped until requested, and mapping rules are changeable at
    runtime via :meth:`BAMGenomeArray.set_mapping`
    
    Parameters
    ----------
    bamfile : One or more filenames or open :class:`pysam.AlignmentFile`
        Filename(s) of `BAM`_ files, or open :class:`pysam.AlignmentFile`
        to include in array. Note: all `BAM`_ files must be sorted and indexed
        by `samtools`_. 
    
    mapping : func
        :term:`mapping function` that determines how each read alignment is mapped to a 
        count at a genomic position. Examples include mapping reads to their
        fiveprime ends, mapping reads to their threeprime ends, or mapping
        somewhere in the middle. Factories to produce such functions are provided.
        See references below. (Default: :func:`CenterMapFactory`)

    
    Attributes
    ----------
    map_fn : function
        :term:`mapping function` used to convert :term:`read alignments` to :term:`counts`
    
    bamfiles : list
        `BAM`_ files or :term:`read alignments`, as a list of :py:class:`pysam.AlignmentFile`

    
    Notes
    -----
    The alignment data in |BAMGenomeArray| objects is immutable. If you need
    to change the data in-place, the |BAMGenomeArray| can be converted to a
    |GenomeArray| or |SparseGenomeArray| via :meth:`~BAMGenomeArray.to_genome_array`
    """

    def __init__(self, *bamfiles, **kwargs):  #mapping=None):
        """Create a |BAMGenomeArray|
        
        Parameters
        ----------
        bamfile : One or more filenames or open :class:`pysam.AlignmentFile`
            Filename(s) of `BAM`_ files, or open :class:`pysam.AlignmentFile`
            to include in array. Note: all `BAM`_ files must be sorted and indexed
            by `samtools`_. 
        
        mapping : func
            :term:`mapping function` that determines how each read alignment is mapped to a 
            count at a genomic position. Examples include mapping reads to their
            fiveprime ends, mapping reads to their threeprime ends, or mapping
            somewhere in the middle. Factories to produce such functions are provided.
            See references below. (Default: :func:`CenterMapFactory`)

        See also
        --------
        plastid.genomics.map_factories.FivePrimeMapFactory
            map reads to 5' ends, with or without applying an offset
        
        plastid.genomics.map_factories.VariableFivePrimeMapFactory
            map reads to 5' ends, choosing an offset determined by read length
        
        plastid.genomics.map_factories.ThreePrimeMapFactory
            map reads to 3' ends, with or without applying an offset
        
        plastid.genomics.map_factories.CenterMapFactory
            map each read fractionally to every position in the read, optionally trimming positions from the ends first
        """
        if len(bamfiles) == 1 and isinstance(bamfiles[0], list):
            bamfiles = bamfiles[0]

        bamfiles = list(multiopen(bamfiles, fn=pysam.AlignmentFile, args=("rb", )))
        #bamfiles = [pysam.AlignmentFile(X,"rb") if isinstance(X,str) else X for X in bamfiles]
        self.bamfiles = bamfiles
        self.map_fn = kwargs.get("mapping", CenterMapFactory())  # if mapping is None else mapping
        self._strands = ("+", "-", ".")
        self._normalize = False

        self._chr_lengths = {}
        for bamfile in self.bamfiles:
            for k, v in zip(bamfile.references, bamfile.lengths):
                self._chr_lengths[k] = max(self._chr_lengths.get(k, 0), v)

        self._chroms = sorted(list(self._chr_lengths.keys()))

        self._filters = OrderedDict()
        self._update()

    def __del__(self):
        for bamfile in self.bamfiles:
            bamfile.close()

    def reset_sum(self):
        """Reset the sum to the total number of mapped reads in the |BAMGenomeArray|
        
        Notes
        -----
        Filters are not applied in this summation. See :meth:`BAMGenomeArray.set_sum`
        to manually set a sum, which e.g. could correspond to a total number of filtered
        reads.
        """
        self._sum = sum([X.mapped for X in self.bamfiles])

    def _update(self):
        """Updates mapping function to suit mapping rules
        """
        self.reset_sum()

    def add_filter(self, name, func):
        """Apply a function to filter reads retrieved from regions before mapping and counting
        
        Parameters
        ----------
        name : str
            A name for the filter. If not unique, will overwrite
            previous filter
        
        func : func
            Filter function. Function must take a
            :class:`pysam.AlignedSegment` as a single parameter, and return
            `True` if that read should be included in output, or `False` otherwise
        
        Notes
        -----
        In Python, `lambda` functions do NOT have their own scope! We strongly
        recomend defining filter functions using the ``def`` syntax to avoid
        namespace collisions.
        
        See also
        --------
        plastid.genomics.map_factories.SizeFilterFactory
            generate filter functions that gate read alignments on size
        """
        self._filters[name] = func

    def remove_filter(self, name):
        """Remove a generic filter
        
        Parameters
        ----------
        name : str
            A name for the filter
        
        Returns
        -------
        func
            the removed filter function
        """
        retval = self._filters.pop(name)
        return retval

    def chroms(self):
        """Returns a list of chromosomes
        
        Returns
        -------
        list
            sorted chromosome names as strings
        """
        return self._chroms

    def lengths(self):
        """Returns a dictionary mapping chromosome names to lengths. 
        
        Returns
        -------
        dict
            mapping chromosome names to lengths
        """
        return self._chr_lengths

    def get_reads_and_counts(self, roi, roi_order=True):
        """Return :term:`read alignments` covering a |GenomicSegment|, and a
        count vector mapping reads to each positions in the |GenomicSegment|,
        following the rule specified by :meth:`~BAMGenomeArray.set_mapping`.
        Reads are strand-matched to `roi` by default. To obtain unstranded reads,
        set the value of `roi.strand` to `'.'`
        
        Parameters
        ----------
        roi : |GenomicSegment|
            Region of interest

        
        Returns
        -------
        list
            List of reads (as :class:`pysam.AlignedSegment`)
            covering region of interest

        numpy.ndarray
            Counts at each position of `roi`, under whatever mapping rule
            was set by :py:meth:`~BAMGenomeArray.set_mapping`


        Raises
        ------
        ValueError
            if bamfiles not sorted or not indexed
        """
        # fetch reads
        chrom = roi.chrom
        strand = roi.strand
        start = roi.start
        end = roi.end

        if chrom not in self.chroms():
            # FIXME: generalize to N-D
            shape = [1] + getattr(self.map_fn, "shape", [])
            return [], numpy.zeros(shape)

        reads = itertools.chain.from_iterable(
            (
                X.fetch(
                    reference=chrom,
                    start=start,
                    end=end,
                    # until_eof=True, # this could speed things up. need to test/investigate
                ) for X in self.bamfiles
            )
        )

        # filter by strand
        if strand == "+":
            reads = ifilter(lambda x: x.is_reverse is False, reads)
        elif strand == "-":
            reads = ifilter(lambda x: x.is_reverse is True, reads)

        # Pass through additional filters (e.g. size filters, if they have
        # been added)
        for my_filter in self._filters.values():
            reads = ifilter(my_filter, reads)

        # retrieve selected parts of regions
        reads, count_array = self.map_fn(list(reads), roi)

        # normalize to reads per million if normalization flag is set
        if self._normalize is True:
            count_array = count_array / float(self.sum()) * 1e6

        if roi_order == True and strand == "-":
            count_array = count_array[..., ::-1]

        return reads, count_array

    def get_reads(self, roi):
        """Returns reads covering a |GenomicSegment|. Reads are strand-matched
        to `roi` by default, and are included or excluded depending upon the
        rule specified by :meth:`~BAMGenomeArray.set_mapping`. To obtain unstranded
        reads, set the value of `roi.strand` to `'.'`
        
        Parameters
        ----------
        roi : |GenomicSegment|
            Region of interest

        
        Returns
        -------
        list
            List of reads (as :class:`pysam.AlignedSegment`) covering region of interest,
            according to the rule set in :meth:`~BAMGenomeArray.set_mapping`


        Raises
        ------
        ValueError
            if bamfiles are not sorted or not indexed
        """
        reads, _ = self.get_reads_and_counts(roi)
        return reads

    def __getitem__(self, roi):
        """Retrieve array of counts from a region of interest, following
        the mapping rule set by :meth:`~BAMGenomeArray.set_mapping`.
        Values in the vector are ordered 5' to 3' relative to `roi`
        rather than the genome (i.e. are reversed for reverse-strand
        features).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome

        Raises
        ------
        ValueError
            if bamfiles not sorted or not indexed

        Returns
        -------
        numpy.ndarray
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        return self.get(roi, roi_order=True)

    def get(self, roi, roi_order=True):
        """Retrieve array of counts from a region of interest, following
        the mapping rule set by :meth:`~BAMGenomeArray.set_mapping`.
        Values in the vector are ordered 5' to 3' relative to `roi`
        rather than the genome (i.e. are reversed for reverse-strand
        features).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome
        
        roi_order : bool, optional
            If `True` (default) return vector of values 5' to 3' 
            relative to vector rather than genome.

        Raises
        ------
        ValueError
            if bamfiles not sorted or not indexed

        Returns
        -------
        numpy.ndarray
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        if isinstance(roi, SegmentChain):
            return roi.get_counts(self)

        _, count_array = self.get_reads_and_counts(roi, roi_order=roi_order)

        return count_array

    def get_mapping(self):
        """Return the docstring of the current mapping function
        """
        return self.map_fn.__doc__

    def set_mapping(self, mapping_function):
        """Change the mapping rule
        
        Parameters
        ----------
        mapping : func
            Function that determines how each read alignment is mapped to a 
            count at a position in the |BAMGenomeArray|. Examples include
            mapping reads to their fiveprime or threeprime ends, with or
            without offsets. Factories to produce such functions are provided.
            See references below.

            
        See also
        --------
        plastid.genomics.map_factories.FivePrimeMapFactory
            map reads to 5' ends, with or without applying an offset
        
        plastid.genomics.map_factories.VariableFivePrimeMapFactory
            map reads to 5' ends, choosing an offset determined by read length
        
        plastid.genomics.map_factories.ThreePrimeMapFactory
            map reads to 3' ends, with or without applying an offset
        
        plastid.genomics.map_factories.CenterMapFactory
            map each read fractionally to every position in the read, optionally trimming positions from the ends first
        """
        self.map_fn = mapping_function
        self._update()

    def to_genome_array(self, array_type=None):
        """Converts |BAMGenomeArray| to a |GenomeArray| or |SparseGenomeArray|
        under the mapping rule set by :meth:`~BAMGenomeArray.set_mapping`

        Parameters
        ----------
        array_type : class
            Type of GenomeArray to return. |GenomeArray| or |SparseGenomeArray|.
            (Default: |GenomeArray|)
        
        Returns
        -------
        |GenomeArray| or |SparseGenomeArray|
        """
        if array_type is None:
            array_type = GenomeArray

        ga = array_type(chr_lengths=self.lengths(), strands=self.strands())
        for chrom in self.chroms():
            for strand in self.strands():
                seg = GenomicSegment(chrom, 0, self.lengths()[chrom] - 1, strand)
                ga[seg] = self[seg]

        return ga

    def to_variable_step(self, fh, trackname, strand, window_size=100000, printer=None, **kwargs):
        """Write the contents of the |BAMGenomeArray| to a variableStep `Wiggle`_ file
        under the mapping rule set by :meth:`~BAMGenomeArray.set_mapping`.
        
        See the `Wiggle spec <http://genome.ucsc.edu/goldenpath/help/wiggle.html>`_ for format details.
        
        Parameters
        ----------
        fh : file-like
            Filehandle to write to
        
        trackname : str
            Name of browser track
          
        strand : str
            Strand of |BAMGenomeArray| to export. `'+'`, `'-'`, or `'.'`
        
        window_size : int
            Size of chromosome/contig to process at a time.
            Larger values are faster but less memory-efficient
        
        printer : file-like, optional
            Something implementing a write() method for output

        **kwargs
            Any other key-value pairs to include in track definition line
        """
        assert strand in self.strands()
        printer = NullWriter() if printer is None else printer
        fh.write("track type=wiggle_0 name=%s" % trackname)
        if kwargs is not None:
            for k, v in sorted(kwargs.items(), key=lambda x: x[0]):
                fh.write(" %s=%s" % (k, v))
        fh.write("\n")

        for chrom in sorted(self.chroms()):
            my_size = self.lengths()[chrom]
            printer.write("Writing chromosome %s..." % chrom)
            fh.write("variableStep chrom=%s span=1\n" % chrom)
            window_starts = xrange(0, self._chr_lengths[chrom], window_size)
            for my_start in window_starts:
                my_end = min(my_start + window_size, my_size)
                my_counts = self.get(
                    GenomicSegment(chrom, my_start, my_end, strand), roi_order=False
                )
                if my_counts.sum() > 0:
                    for idx in my_counts.nonzero()[0]:
                        genomic_x = my_start + idx
                        val = my_counts[idx]
                        fh.write("%s\t%s\n" % (genomic_x + 1, val))

    def to_bedgraph(self, fh, trackname, strand, window_size=100000, printer=None, **kwargs):
        """Write the contents of the |BAMGenomeArray| to a `bedGraph`_ file
        under the mapping rule set by :meth:`~BAMGenomeArray.set_mapping`.
        
        See the `bedGraph spec <https://cgwb.nci.nih.gov/goldenPath/help/bedgraph.html>`_
        for format details.
            
        Parameters
        ----------
        fh : file-like
            Filehandle to write to
        
        trackname : str
            Name of browser track
          
        strand : str
            Strand of |BAMGenomeArray| to export. `'+'`, `'-'`, or `'.'`
        
        window_size : int
            Size of chromosome/contig to process at a time.
            Larger values are faster but less memory-efficient
         
        printer : file-like, optional
            Something implementing a write() method for output

        **kwargs
            Any other key-value pairs to include in track definition line
        """
        assert strand in self.strands()
        assert window_size > 0
        printer = NullWriter() if printer is None else printer
        # write header
        fh.write("track type=bedGraph name=%s" % trackname)
        if kwargs is not None:
            for k, v in sorted(kwargs.items(), key=lambda x: x[0]):
                fh.write(" %s=%s" % (k, v))
        fh.write("\n")

        for chrom in sorted(self.chroms()):
            printer.write("Writing chromosome %s..." % chrom)
            window_starts = xrange(0, self._chr_lengths[chrom], window_size)
            my_size = self.lengths()[chrom]
            for my_start in window_starts:
                my_end = min(my_start + window_size, my_size)
                my_counts = self.get(
                    GenomicSegment(chrom, my_start, my_end, strand),
                    roi_order=False
                )

                if my_counts.sum() > 0:
                    genomic_start_x = my_start
                    last_val = my_counts[0]

                    for x, val in enumerate(my_counts[1:]):
                        if val != last_val:
                            genomic_end_x = 1 + x + my_start
                            #write line: chrom chromStart chromEnd dataValue. 0-based half-open
                            if last_val > 0:
                                fh.write(
                                    "%s\t%s\t%s\t%s\n" %
                                    (chrom, genomic_start_x, genomic_end_x, last_val)
                                )

                            #update variables
                            last_val = val
                            genomic_start_x = genomic_end_x
                        else:
                            continue
                    # write out last values for window
                    if last_val > 0:
                        fh.write("%s\t%s\t%s\t%s\n" % (chrom, genomic_start_x, my_end, last_val))


class BigWigGenomeArray(AbstractGenomeArray):
    """BigWigGenomeArray(maxmem=0)
    
    High-performance GenomeArray for `BigWig`_ files.

    Parameters
    ----------
    maxmem : float
        Maximum desired memory footprint for C objects, in megabytes.
        May be temporarily exceeded if large queries are requested.
        (Default: 0, No maximum)        
    
    """

    def __init__(self, maxmem=0, **kwargs):
        """Create a |BigWigGenomeArray|.
        
        `BigWig`_ files may be added to the array via
        :meth:`~BigWigGenomeArray.add_from_bigwig`.
        
        Parameters
        ----------
        maxmem : float
            Maximum desired memory footprint for C objects, in megabytes.
            May be temporarily exceeded if large queries are requested.
            (Default: 0, No maximum)        
        """
        self._strand_dict = {}
        self._normalize = False
        self._chromset = None
        self._sum = None
        self._lengths = None
        self._strands = []
        self._maxmem = maxmem
        self.fill = 0.0  # fill

    def __getitem__(self, roi):
        """Retrieve array of counts from a region of interest.
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome

        Returns
        -------
        numpy.ndarray
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        return self.get(roi, roi_order=True)

    def get(self, roi, roi_order=True):
        """Retrieve array of counts from a region of interest.
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome
        
        roi_order : bool, optional
            If `True` (default) return vector of values 5' to 3' 
            relative to vector rather than genome.

        Returns
        -------
        numpy.ndarray
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        if isinstance(roi, SegmentChain):
            return roi.get_counts(self)

        strand = roi.strand
        sdict = self._strand_dict

        count_vec = numpy.zeros(len(roi))
        if strand in sdict:
            for bw in sdict[strand]:
                 # we flip strand later to save operations
                count_vec += bw.get(roi, roi_order=False)
        else:
            warnings.warn(
                "Strand '%s' not in BigWigGenomeArray (has %s)." %
                (strand, ", ".join(sdict.keys())), DataWarning
            )

        if self._normalize is True:
            count_vec = count_vec / float(self.sum()) * 1e6

        if roi_order == True and strand == "-":
            count_vec = count_vec[::-1]

        return count_vec

    def strands(self):
        """Return a tuple of strands in the GenomeArray
        
        Returns
        -------
        tuple
            Chromosome strands as strings
        """
        return self._strands

    def chroms(self):
        """Return a list of chromosomes in the GenomeArray
        
        Returns
        -------
        list
            Chromosome names as strings
        """
        if self._chromset is not None:
            return self._chromset
        else:
            chromset = []
            for ltmp in self._strand_dict.values():
                for bw in ltmp:
                    chromset.extend(bw.chroms.keys())

            self._chromset = set(chromset)
            return self._chromset

    def lengths(self):
        """Return a dictionary mapping chromosome names to lengths.
        When two strands report different lengths for a chromosome, the
        max length is taken.
        
        Returns
        -------
        dict
            Dictionary mapping chromosome names to chromosome lengths
        """
        if self._lengths is not None:
            return self._lengths
        else:
            lengths = {}
            for ltmp in self._strand_dict.values():
                for bw in ltmp:
                    sizedict = bw.chroms
                    for k, v in sizedict.items():
                        lengths[k] = max(lengths.get(k, 0), v)

            self._lengths = lengths
            return lengths

    def add_from_bigwig(self, filename, strand):
        """Import additional data from a `BigWig`_ file
        
        Parameters
        ----------
        filename : str
            Path to `BigWig`_ file
        
        strand : str
            Strand to which data should be added. `'+'`, `'-'`, or `'.'`
        """
        self._chromset = None
        self._lengths = None
        self._sum = None
        bw = BigWigReader(filename, fill=self.fill, maxmem=self._maxmem)

        try:
            self._strand_dict[strand].append(bw)
        except KeyError:
            self._strand_dict[strand] = [bw]

        self._strands = sorted(self._strand_dict.keys())

    def reset_sum(self):
        """Reset sum to total of data in the |BigWigGenomeArray|"""
        my_sum = 0
        for ltmp in self._strand_dict.values():
            for bw in ltmp:
                my_sum += bw.sum()

        self._sum = my_sum
        return my_sum

    def to_genome_array(self):
        """Converts |BigWigGenomeArray| to a |GenomeArray|

        Returns
        -------
        |GenomeArray|
        """
        ga = GenomeArray(chr_lengths=self.lengths(), strands=self.strands())
        for chrom, length in self.lengths().items():
            for strand in self._strands:
                region = GenomicSegment(chrom, 0, length, strand)
                for reader in self._strand_dict.get(strand, []):
                    old = ga.get(region, roi_order=False)
                    new = old + reader.get_chromosome_counts(chrom)
                    ga.__setitem__(region, new, roi_order=False)

        return ga


class GenomeArray(MutableAbstractGenomeArray):
    """Array-like data structure that maps numerical values (e.g. read :term:`counts`
    or conservation data, et c) to nucleotide positions in a genome.

    Data in the array may be manipulated manually and/or imported from `Wiggle`_
    or `BedGraph`_ files, as well as `bowtie`_ alignments.
    
    |GenomeArray| supports basic mathematical operations elementwise, where elements
    are nucleotide positions.
    
    Parameters
    ----------
    chr_lengths : dict or `None`, optional
        Dictionary mapping chromosome names to lengths.
        Suppyling this in advance yields considerable
        speed improvements, as memory can be pre-allocated
        for each chromosomal position. If not provided,
        a minimum chromosome size will be guessed, and
        chromosomes re-sized as needed.
    
    min_chr_size : int
        If chr_lengths is not supplied, `min_chr_size` is 
        the default first guess of a chromosome size. If
        the genome has large chromosomes, it is much
        much better, speed-wise, to provide `chr_lengths`
        than to provide a guess here that is too small.
    
    strands : sequence
        Sequence of strand names for the |GenomeArray|. (Default: `('+','-')`)    
    """

    def __init__(self, chr_lengths=None, strands=None, min_chr_size=MIN_CHR_SIZE):
        """Create a |GenomeArray|
        
        Parameters
        ----------
        chr_lengths : dict or `None`, optional
            Dictionary mapping chromosome names to lengths.
            Suppyling this in advance yields considerable
            speed improvements, as memory can be pre-allocated
            for each chromosomal position. If not provided,
            a minimum chromosome size will be guessed, and
            chromosomes re-sized as needed.
        
        min_chr_size : int
            If chr_lengths is not supplied, `min_chr_size` is 
            the default first guess of a chromosome size. If
            the genome has large chromosomes, it is much
            much better, speed-wise, to provide `chr_lengths`
            than to provide a guess here that is too small.
        
        strands : sequence
            Sequence of strand names for the |GenomeArray|. (Default: `('+','-')`)
        """
        self._chroms = {}
        self._strands = _DEFAULT_STRANDS if strands is None else strands
        self.min_chr_size = min_chr_size
        self._sum = None
        self._normalize = False
        if chr_lengths is not None:
            for chrom in chr_lengths.keys():
                self._chroms[chrom] = {}
                for strand in self._strands:
                    l = chr_lengths[chrom]
                    self._chroms[chrom][strand] = numpy.zeros(l)

    def reset_sum(self):
        """Reset the sum of the |GenomeArray| to the sum of all positions in the array
        """
        self._sum = sum([X.sum() for X in self.iterchroms()])

    def _has_same_dimensions(self, other):
        """Return `True` if `self` and `other` have the chromosomes, strands, and chromosome lengths
        
        Parameters
        ----------
        other : |GenomeArray|
        
        
        Returns
        -------
        bool
        """
        assert set(self.keys()) == set(other.keys())
        assert self.strands() == other.strands()
        assert self.lengths() == other.lengths()

    def __eq__(self, other, tol=1e-10):
        """Test equality between `self` and `other`.
        
        To be equal, both |GenomeArrays| must have identical values at all
        nonzero positions in either array. Chromosome lengths need not be equal. 
        
        Parameters
        ----------
        other : |GenomeArray| or |SparseGenomeArray|
        
        
        Returns
        -------
        bool
        """
        snz = self.nonzero()
        onz = other.nonzero()
        for chrom in set(self.chroms()) | set(other.chroms()):
            for strand in set(self.strands()) | set(other.strands()):
                self_nonzero_vec = snz.get(chrom,
                                           {K: numpy.array([])
                                            for K in self.strands()}).get(strand, numpy.array([]))
                other_nonzero_vec = onz.get(chrom, {K: numpy.array([])
                                                    for K in other.strands()
                                                    }).get(strand, numpy.array([]))

                if len(self_nonzero_vec) != len(other_nonzero_vec):
                    return False

                # indices of nonzero positions must be same
                if (self_nonzero_vec != other_nonzero_vec).any():
                    return False

                # values of nonzero positions must be same
                if len(self_nonzero_vec) > 0:
                    test_seg = GenomicSegment(
                        chrom, self_nonzero_vec.min(), self_nonzero_vec.max(), strand
                    )
                    if not (self[test_seg] - other[test_seg] <= tol).all():
                        return False

        return True

    def __getitem__(self, roi):
        """Retrieve array of counts from a region of interest (`roi`)
        with values in vector ordered 5' to 3' relative to `roi`
        rather than genome (i.e. are reversed for reverse-strand
        features).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome
        
        Returns
        -------
        :class:`numpy.ndarray`
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        return self.get(roi, roi_order=True)

    def get(self, roi, roi_order=True):
        """Retrieve array of counts from a region of interest (`roi`)
        with values in vector ordered 5' to 3' relative to `roi`
        rather than genome (i.e. are reversed for reverse-strand
        features).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome
        
        roi_order : bool, optional
            If `True` (default) return vector of values 5' to 3' 
            relative to vector rather than genome.

        Returns
        -------
        :class:`numpy.ndarray`
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        plastid.genomics.roitools.SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        if isinstance(roi, SegmentChain):
            return roi.get_counts(self)

        chrom = roi.chrom
        strand = roi.strand
        start = roi.start
        end = roi.end
        try:
            assert roi.end < len(self._chroms[chrom][strand])
        except AssertionError:
            my_len = len(self._chroms[chrom][strand])
            new_size = max(my_len + 10000, end + 10000)
            for my_strand in self.strands():
                new_strand = copy.deepcopy(self._chroms[roi.chrom][my_strand])
                new_strand.resize(new_size)
                self._chroms[chrom][my_strand] = new_strand
        except KeyError:
            assert strand in self.strands()
            if chrom not in self.keys():
                self._chroms[chrom] = {}
                for my_strand in self.strands():
                    self._chroms[chrom][my_strand] = numpy.zeros(self.min_chr_size)

        vals = self._chroms[chrom][strand][start:end]
        if self._normalize is True:
            vals = 1e6 * vals / self.sum()

        if roi_order == True and strand == "-":
            vals = vals[::-1]

        return vals

    def __setitem__(self, seg, val, roi_order=True):
        """Set values in the |GenomeArray| over a region of interest.
        
        If the `seg` is outside the bounds of the current |GenomeArray|,
        the |GenomeArray| will be automatically expanded to accomodate
        the genomic coordinates of `seg`.

        Parameters
        ----------
        seg : |GenomicSegment| or |SegmentChain|
            Region of interest

        val : int, float, or :class:`numpy.ndarray`
            Scalar or vector of values to set in array over `seg`.
            If a vector, values should be ordered 5'-to-3' relative
            to `seg` rather than (i.e. for a reverse-strand feature,
            position 0 in the vector would correspond to seg.end)
            in the genome

        roi_order : bool, optional
            If `True` (default) and `val` is a vector, values in `val`
            are assorted to go 5' to 3' relative to `seg` rather than
            genome
        """
        self._sum = None

        if isinstance(seg, SegmentChain):
            if isinstance(val, numpy.ndarray):
                if seg.spanning_segment.strand == "-":
                    val = val[::-1]

            x = 0
            for subseg in seg:
                if isinstance(val, numpy.ndarray):
                    subval = val[x:x + len(subseg)]
                else:
                    subval = val
                x += len(subseg)
                self.__setitem__(subseg, subval, roi_order=False)

            return

        old_normalize = self._normalize
        if old_normalize == True:
            warn(
                "Temporarily turning off normalization during value set. It will be re-enabled automatically when complete.",
                DataWarning
            )

        self.set_normalize(False)

        chrom = seg.chrom
        strand = seg.strand
        start = seg.start
        end = seg.end
        if strand == "-" and isinstance(val, numpy.ndarray) and roi_order == True:
            val = val[::-1]

        try:
            assert end < len(self._chroms[seg.chrom][seg.strand])
        except AssertionError:
            my_len = len(self._chroms[chrom][strand])
            new_size = max(my_len + 10000, end + 10000)
            for my_strand in self.strands():
                # copy then resize, because resize() can't work in-place with refs to array
                new_strand = copy.deepcopy(self._chroms[chrom][my_strand])
                new_strand.resize(new_size)
                self._chroms[chrom][my_strand] = new_strand
        except KeyError:
            assert strand in self.strands()
            if chrom not in self.keys():
                self._chroms[chrom] = {}
                for my_strand in self.strands():
                    self._chroms[chrom][my_strand] = numpy.zeros(self.min_chr_size)

        self._chroms[chrom][strand][start:end] = val
        self.set_normalize(old_normalize)

    def keys(self):
        """Return names of chromosomes in the GenomeArray"""
        return self.chroms()

    def iterchroms(self):
        """Return an iterator of each chromosome strand array
        
        Yields
        ------
        numpy.ndarray
            Positionwise counts on a chromosome strand
        """
        for chrom in self.keys():
            for strand in self.strands():
                yield self._chroms[chrom][strand]

    #TODO: no unit test for this
    def plot(self, chroms=None, strands=None, **plot_kw):
        """Create a plot of coverage along tracks or chromosomes
        
        Parameters
        ----------
        chroms : list
            Chromosomes to plot (default: all)
            
        strands : list
            Strands to plot (default: all)
            
        plot_kw
            A dictionary of keywords to pass to matplotlib
        
        
        Returns
        -------
        :py:class:`matplotlib.figure.Figure`
        """
        import matplotlib
        matplotlib.use("Agg")
        import matplotlib.pyplot as plt

        fig_kw = {'figsize': (8, 10)}
        colors = {"+": "blue", "-": "orange", ".": "darkgreen"}

        if chroms is None:
            chroms = self.keys()

        if strands is None:
            strands = self.strands()

        fig_kw.update(plot_kw)
        f, axes = plt.subplots(nrows=len(chroms), **fig_kw)
        plt.subplots_adjust(hspace=0.3)
        for n, chrom in enumerate(sorted(chroms)):
            x = numpy.arange(0, self.lengths()[chrom])
            for strand in strands:
                multiplier = -1 if strand == "-" else 1
                axes[n].plot(
                    x, self._chroms[chrom][strand] * multiplier, label=strand, color=colors[strand]
                )
                axes[n].set_title(chrom)
                axes[n].set_ylabel("Counts")
                axes[n].set_xlabel("Position (nt)")
                axes[n].ticklabel_format(useOffset=1, style='plain', axis='x')

        return f

    def nonzero(self):
        """Return the indices of chromosomal positions with non-zero values
        at each chromosome/strand pair. Results are returned as a hierarchical
        dictionary mapping chromosome names, to a dictionary of strands,
        which in turn map to  to arrays of non-zero indices on that chromosome-strand.
        
        Returns
        -------
        dict
            `dict[chrom][strand]` = numpy.ndarray of indices
        """
        d_out = {}
        for key in self.keys():
            d_out[key] = {}
            for strand in self.strands():
                d_out[key][strand] = self._chroms[key][strand].nonzero()[0]
        return d_out

    def apply_operation(self, other, func, mode="same"):
        """Applies a binary operator to a copy of `self` and to `other` elementwise.
        `other` may be a scalar quantity or another |GenomeArray|. In both cases,
        a new |GenomeArray| is returned, and `self` is left unmodified.
        If :meth:`set_normalize` is set to `True`, it is disabled during the
        operation.
        
        Parameters
        ----------
        other : float, int, or |MutableAbstractGenomeArray|
            Second argument to `func`
        
        func : func
            Function to perform. This must take two arguments.
            a :py:class:`numpy.ndarray` (chromosome-strand) from the
            |GenomeArray| will be supplied as the first argument, and the
            corresponding chromosome-strand from `other` as the second.
            This operation will be applied over all chromosomes and strands.
            
            If mode is set to `all`, and a chromosome or strand is not
            present in one of self or other, zero will be supplied
            as the missing argument to func. func must handle this
            gracefully.
        
        mode : str, choice of `'same'`, `'all'`, or `'truncate'`
            Only relevant if `other` is a |MutableAbstractGenomeArray|
        
            If '`same`' each set of corresponding chromosomes must
            have the same dimensions in both parent |GenomeArray| s.
            In addition, both |GenomeArray| s must have the same
            chromosomes, and the same strands.

            If '`all`', corresponding chromosomes in the parent
            |GenomeArray| s will be resized to the dimensions of
            the larger chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; all chromosomes and strands will be
            included in output. 
                    
            If `'truncate'`,corresponding chromosomes in the parent
            |GenomeArray| s will be resized to the dimensions of
            the smaller chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; chromosomes or strands not common
            to both parents will be ignored.    
        
        Returns
        -------
        |GenomeArray|
            new |GenomeArray| after the operation is applied
        """
        new_array = GenomeArray.like(self)
        old_normalize = self._normalize
        if old_normalize == True:
            warn(
                "Temporarily turning off normalization during value set. It will be re-enabled automatically when complete.",
                DataWarning
            )

        if type(other) == self.__class__:
            if mode == "same":
                self._has_same_dimensions(other)
                for chrom in self.keys():
                    for strand in self.strands():
                        seg = GenomicSegment(chrom, 0, len(self._chroms[chrom][strand]), strand)
                        new_array[seg] = func(self[seg], other[seg])
            elif mode == "all":
                chroms = {}.fromkeys(set(self.keys()) | set(other.keys()))
                strands = set(self.strands()) | set(other.strands())
                new_array = GenomeArray(chroms, strands=strands)
                for chrom in chroms:
                    if chrom in self.keys() and chrom in other.keys():
                        for strand in strands:
                            if strand in self.strands() and strand in other.strands():
                                sc = copy.deepcopy(self._chroms[chrom][strand])
                                oc = copy.deepcopy(other._chroms[chrom][strand])
                                if len(sc) > len(oc):
                                    oc.resize(len(sc), refcheck=False)
                                elif len(oc) > len(sc):
                                    sc.resize(len(oc), refcheck=False)
                                new_array._chroms[chrom][strand] = func(sc, oc)
                            elif strand in self.strands():
                                new_array._chroms[chrom][strand] = func(
                                    copy.deepcopy(self._chroms[chrom][strand]), 0
                                )
                            else:
                                new_array._chroms[chrom][strand] = func(
                                    copy.deepcopy(other._chroms[chrom][strand]), 0
                                )
                    elif chrom in self.keys():
                        for strand in strands:
                            new_array._chroms[chrom][strand] = func(
                                copy.deepcopy(self[chrom][strand]), 0
                            )
                    else:
                        for strand in strands:
                            new_array._chroms[chrom][strand] = func(
                                copy.deepcopy(other[chrom][strand]), 0
                            )
            elif mode == "truncate":
                my_strands = set(self.strands()) & set(other.strands())
                my_chroms = set(self.chroms()) & set(other.chroms())
                for chrom in my_chroms:
                    for strand in my_strands:
                        sc = copy.deepcopy(self._chroms[chrom][strand])
                        oc = copy.deepcopy(other._chroms[chrom][strand])
                        if len(sc) > len(oc):
                            sc.resize(len(oc), refcheck=False)
                        elif len(oc) > len(sc):
                            oc.resize(len(sc), refcheck=False)
                        new_array._chroms[chrom][strand] = func(sc, oc)
            else:
                raise ValueError("Mode not understood. Must be 'same', 'all', or 'truncate.'")
        else:
            for chrom in self.keys():
                for strand in self.strands():
                    new_array._chroms[chrom][strand] = func(self._chroms[chrom][strand], other)

        self.set_normalize(old_normalize)
        return new_array

    def __add__(self, other, mode="all"):
        """Add `other` to `self`, returning a new |GenomeArray|
        and leaving `self` unchanged. `other` may be a scalar quantity or
        another |GenomeArray|. In the latter case, addition is elementwise.
        
        Parameters
        ----------
        other : float, int, or |MutableAbstractGenomeArray|
        
        mode : str, choice of `'same'`, `'all'`, or `'truncate'`
            Only relevant if `other` is a |MutableAbstractGenomeArray|
        
            If '`same`' each set of corresponding chromosomes must
            have the same dimensions in both parent |GenomeArrays|.
            In addition, both |GenomeArrays| must have the same
            chromosomes, and the same strands.

            If '`all`', corresponding chromosomes in the parent
            |GenomeArrays| will be resized to the dimensions of
            the larger chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; all chromosomes and strands will be
            included in output. 
                    
            If `'truncate'`,corresponding chromosomes in the parent
            |GenomeArrays| will be resized to the dimensions of
            the smaller chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; chromosomes or strands not common
            to both parents will be ignored.     
        
        Returns
        -------
        |GenomeArray|
        """
        return self.apply_operation(other, operator.add, mode=mode)

    def __mul__(self, other, mode="all"):
        """Multiply `other` by `self`, returning a new |GenomeArray|
        and leaving `self` unchanged. `other` may be a scalar quantity or
        another |GenomeArray|. In the latter case, multiplication is elementwise.
        
        Parameters
        ----------
        other : float, int, or |MutableAbstractGenomeArray|
        
        mode : str, choice of `'same'`, `'all'`, or `'truncate'`
            Only relevant if `other` is a |MutableAbstractGenomeArray|
        
            If '`same`' each set of corresponding chromosomes must
            have the same dimensions in both parent |GenomeArrays|.
            In addition, both |GenomeArrays| s must have the same
            chromosomes, and the same strands.

            If '`all`', corresponding chromosomes in the parent
            |GenomeArrays| will be resized to the dimensions of
            the larger chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; all chromosomes and strands will be
            included in output. 
                    
            If `'truncate'`,corresponding chromosomes in the parent
            |GenomeArrays| will be resized to the dimensions of
            the smaller chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; chromosomes or strands not common
            to both parents will be ignored.                 
        
        Returns
        -------
        |GenomeArray|
        """
        return self.apply_operation(other, operator.mul, mode=mode)

    def __sub__(self, other, mode="all"):
        """Subtract `other` from `self`, returning a new |GenomeArray|
        and leaving `self` unchanged. `other` may be a scalar quantity or
        another |GenomeArray|. In the latter case, subtraction is elementwise.
        
        Parameters
        ----------
        other : float, int, or |MutableAbstractGenomeArray|
        
        mode : str, choice of `'same'`, `'all'`, or `'truncate'`
            Only relevant if `other` is a |MutableAbstractGenomeArray|
        
            If '`same`' each set of corresponding chromosomes must
            have the same dimensions in both parent |GenomeArrays|.
            In addition, both |GenomeArrays| s must have the same
            chromosomes, and the same strands.

            If '`all`', corresponding chromosomes in the parent
            |GenomeArrays| will be resized to the dimensions of
            the larger chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; all chromosomes and strands will be
            included in output. 
                    
            If `'truncate'`,corresponding chromosomes in the parent
            |GenomeArrays| will be resized to the dimensions of
            the smaller chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; chromosomes or strands not common
            to both parents will be ignored.                 
        
        Returns
        -------
        |GenomeArray|
        """
        return self.__add__(other * -1)

    def add_from_bowtie(self, fh, mapfunc, min_length=25, max_length=numpy.inf, **trans_args):
        """Import alignment data in native `bowtie`_ format to the current GenomeArray
        
        Parameters
        ----------
        fh : file-like
            filehandle pointing to `bowtie`_ file
        
        mapfunc : func
            a :term:`mapping function` that transforms alignment
            coordinates into a sequence of (|GenomicSegment|,value) pairs

        min_length : int, optional
            minimum length for read to be counted
            (Default: 25)
        
        max_length : int or numpy.inf, optional
            maximum length for read to be counted
            (Default: infinity)
        
        trans_args : dict, optional
            Keyword arguments to pass to transformation
            function

            
        See also
        --------
        five_prime_map
            map reads to 5' ends, with or without applying an offset
        
        variable_five_prime_map
            map reads to 5' ends, choosing an offset determined by read length
        
        three_prime_map
            map reads to 3' ends, with or without applying an offset
        
        center_map
            map each read fractionally to every position in the read, optionally
            trimming positions from the ends first            
        """
        for feature in BowtieReader(fh):
            span_len = len(feature.spanning_segment)
            if span_len >= min_length and span_len <= max_length:
                tuples = mapfunc(feature, **trans_args)
                for seg, val in tuples:
                    self[seg] += val

        self._sum = None

    def add_from_wiggle(self, fh, strand):
        """Import data from a `Wiggle`_ or `bedGraph`_ file to current GenomeArray
        
        Parameters
        ----------
        fh : file-like
            filehandle pointing to wiggle file
        
        strand : str
            Strand to which data should be added. `'+'`, `'-'`, or `'.'`
        """
        assert strand in self.strands()
        for chrom, start, stop, val in WiggleReader(fh):
            seg = GenomicSegment(chrom, start, stop, strand)
            self[seg] += val

        self._sum = None

    def to_variable_step(self, fh, trackname, strand, printer=None, **kwargs):
        """Export the contents of the GenomeArray to a variable step
        `Wiggle`_ file. For sparse data, `bedGraph`_ can be more efficient format.
        
        See the `Wiggle spec <http://genome.ucsc.edu/goldenpath/help/wiggle.html>`_
        for format details
        
        Parameters
        ----------
        fh : file-like
            Open filehandle to which data will be written
        
        trackname : str
            Name of browser track
        
        strand : str
            Strand to export. `'+'`, `'-'`, or `'.'`
        
        printer : file-like, optional
            Something implementing a write() method for output
            
        **kwargs
            Any other key-value pairs to include in track definition line
        """
        assert strand in self.strands()
        printer = NullWriter() if printer is None else printer
        fh.write("track type=wiggle_0 name=%s" % trackname)
        if kwargs is not None:
            for k, v in sorted(kwargs.items(), key=lambda x: x[0]):
                fh.write(" %s=%s" % (k, v))
        fh.write("\n")
        nonzero = self.nonzero()
        for chrom in sorted(nonzero):
            printer.write("Writing chromosome %s..." % chrom)
            fh.write("variableStep chrom=%s span=1\n" % chrom)
            indices = nonzero[chrom][strand]
            for idx in indices:
                val = self._chroms[chrom][strand][self._slicewrap(idx)]
                fh.write("%s\t%s\n" % (idx + 1, val))

    def to_bedgraph(self, fh, trackname, strand, printer=None, **kwargs):
        """Write the contents of the GenomeArray to a `bedGraph`_ file
        
        See the `bedGraph spec <https://cgwb.nci.nih.gov/goldenPath/help/bedgraph.html>`_
        for format details
            
        Parameters
        ----------
        fh : file-like
            Open filehandle to which data will be written
        
        trackname : str
            Name of browser track
        
        strand : str
            Strand to export. `'+'`, `'-'`, or `'.'`

        printer : file-like, optional
            Something implementing a write() method for output       

        **kwargs
            Any other key-value pairs to include in track definition line
        """
        assert strand in self.strands()
        printer = NullWriter() if printer is None else printer
        fh.write("track type=bedGraph name=%s" % trackname)
        if kwargs is not None:
            for k, v in sorted(kwargs.items(), key=lambda x: x[0]):
                fh.write(" %s=%s" % (k, v))
        fh.write("\n")
        nonzero = self.nonzero()
        for chrom in sorted(nonzero):
            printer.write("Writing chromosome %s..." % chrom)
            if self._chroms[chrom][strand].sum() > 0:
                last_val = 0
                last_x = 0
                nz = nonzero[chrom][strand]
                for x in range(nz.min(), nz.max() + 1):
                    val = self._chroms[chrom][strand][self._slicewrap(x)]
                    if val != last_val:
                        # line is: chrom chromStart chromEnd dataValue
                        fh.write("%s\t%s\t%s\t%s\n" % (chrom, last_x, x, last_val))
                        # prepare for next loop
                        last_val = val
                        last_x = x
                    else:
                        continue
                # write final line, not included in loop
                fh.write("%s\t%s\t%s\t%s\n" % (chrom, last_x, x + 1, last_val))

    @staticmethod
    def like(other):
        """Return a |GenomeArray| of same dimension as the input array
        
        Parameters
        ----------
        other : |GenomeArray|
        
        Returns
        -------
        GenomeArray
            empty |GenomeArray| of same size as `other`
        """
        return GenomeArray(other.lengths(), strands=other.strands())

    def _slicewrap(self, x):
        """Helper function to wrap coordinates for VariableStep/`bedGraph`_ export"""
        return x


class SparseGenomeArray(GenomeArray):
    """A memory-efficient sublcass of |GenomeArray| using sparse internal representation.
    Note, savings in memory may come at a cost in performance when repeatedly
    getting/setting values, compared to a |GenomeArray|.
    

    Parameters
    ----------
    chr_lengths : dict, optional
        Dictionary mapping chromosome names to lengths.
        Suppyling this parameter yields considerable
        speed improvements, as memory can be pre-allocated
        for each chromosomal position. If not provided,
        a minimum chromosome size will be guessed, and
        chromosomes re-sized as needed (Default: `{}` )
    
    min_chr_size : int,optional
        If `chr_lengths` is not supplied, `min_chr_size` is 
        the default first guess of a chromosome size. If
        your genome has large chromosomes, it is much
        much better, speed-wise, to provide a chr_lengths
        dict than to provide a guess here that is too small.
        (Default: %s)
    
    strands : sequence
        Sequence of strand names for the |GenomeArray|. (Default: `('+','-')`)    
    """

    def __init__(self, chr_lengths=None, strands=None, min_chr_size=MIN_CHR_SIZE):
        """Create a |SparseGenomeArray|

        Parameters
        ----------
        chr_lengths : dict, optional
            Dictionary mapping chromosome names to lengths.
            Suppyling this parameter yields considerable
            speed improvements, as memory can be pre-allocated
            for each chromosomal position. If not provided,
            a minimum chromosome size will be guessed, and
            chromosomes re-sized as needed (Default: `{}` )
        
        min_chr_size : int,optional
            If `chr_lengths` is not supplied, `min_chr_size` is 
            the default first guess of a chromosome size. If
            your genome has large chromosomes, it is much
            much better, speed-wise, to provide a chr_lengths
            dict than to provide a guess here that is too small.
            (Default: %s)
        
        strands : sequence
            Sequence of strand names for the |GenomeArray|. (Default: `('+','-')`)
        """ % MIN_CHR_SIZE
        self._chroms = {}
        self._strands = _DEFAULT_STRANDS if strands is None else strands
        self._sum = None
        self._normalize = False
        self.min_chr_size = min_chr_size
        if chr_lengths is not None:
            for chrom in chr_lengths.keys():
                self._chroms[chrom] = {}
                for strand in self._strands:
                    l = chr_lengths[chrom]
                    self._chroms[chrom][strand] = scipy.sparse.dok_matrix((1, l))

    def lengths(self):
        """Return a dictionary mapping chromosome names to lengths. In the
        case where two strands report different lengths for a chromosome, the
        max length is taken.
        
        Returns
        -------
        dict
            mapping chromosome names to lengths
        """
        d_out = {}.fromkeys(self.keys())
        for key in d_out:
            d_out[key] = max([self._chroms[key][X].shape[1] for X in self.strands()])

        return d_out

    def get(self, roi, roi_order=True):
        """Retrieve array of counts from a region of interest (`roi`)
        with values in vector ordered 5' to 3' relative to `roi`
        rather than genome (i.e. are reversed for reverse-strand
        features).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Region of interest in genome
        
        roi_order : bool, optional
            If `True` (default) return vector of values 5' to 3' 
            relative to `roi` rather than genome.

        Returns
        -------
        :class:`numpy.ndarray`
            vector of numbers, each position corresponding to a position
            in `roi`, from 5' to 3' relative to `roi`
        
        See also
        --------
        SegmentChain.get_counts
            Fetch a spliced vector of data covering a |SegmentChain|
        """
        if isinstance(roi, SegmentChain):
            return roi.get_counts(self)

        if roi.chrom not in self:
            self._chroms[roi.chrom] = {
                K: copy.deepcopy(scipy.sparse.dok_matrix((1, self.min_chr_size)))
                for K in self.strands()
            }
        if roi.end > self._chroms[roi.chrom][roi.strand].shape[1]:
            for strand in self.strands():
                self._chroms[roi.chrom][strand].resize((1, roi.end + 10000))

        vals = self._chroms[roi.chrom][roi.strand][0, roi.start:roi.end]
        if self._normalize is True:
            vals = 1e6 * vals / self.sum()

        vals = vals.toarray().reshape(vals.shape[1], )
        if roi.strand == "-" and roi_order == True:
            vals = vals[::-1]

        return vals

    def __setitem__(self, seg, val, roi_order=True):
        """Set values in the |SparseGenomeArray| over a region of interest.
        
        If the `seg` is outside the bounds of the current |GenomeArray|,
        the |GenomeArray| will be automatically expanded to accomodate
        the genomic coordinates of `seg`.

        Parameters
        ----------
        seg : |GenomicSegment| or |SegmentChain|
            Region of interest

        val : int, float, or :class:`numpy.ndarray`
            Scalar or vector of values to set in array over `seg`.
            If a vector, values should be ordered 5'-to-3' relative
            to `seg` rather than (i.e. for a reverse-strand feature,
            position 0 in the vector would correspond to seg.end)
            in the genome

        roi_order : bool, optional
            If `True` (default) and `val` is a vector, values in `val`
            are assorted to go 5' to 3' relative to `seg` rather than
            genome
       """
        self._sum = None

        if isinstance(seg, SegmentChain):
            if isinstance(val, numpy.ndarray):
                if seg.spanning_segment.strand == "-":
                    val = val[::-1]

            x = 0
            for subseg in seg:
                if isinstance(val, numpy.ndarray):
                    subval = val[x:x + len(subseg)]
                else:
                    subval = val
                x += len(subseg)
                self.__setitem__(subseg, subval, roi_order=False)

            return

        old_normalize = self._normalize

        if isinstance(val, numpy.ndarray) and seg.strand == "-" and roi_order == True:
            val = val[::-1]

        if old_normalize == True:
            warn(
                "Temporarily turning off normalization during value set. It will be re-enabled automatically when complete.",
                DataWarning
            )
        self.set_normalize(False)

        if seg.chrom not in self:
            self._chroms[seg.chrom] = {
                K: copy.deepcopy(scipy.sparse.dok_matrix((1, self.min_chr_size)))
                for K in self.strands()
            }
        if seg.end > self._chroms[seg.chrom][seg.strand].shape[1]:
            for strand in self.strands():
                self._chroms[seg.chrom][strand].resize((1, seg.end + 10000))

        self._chroms[seg.chrom][seg.strand][0, seg.start:seg.end] = val
        self.set_normalize(old_normalize)

    def __mul__(self, other, mode=None):
        """Multiply `other` by `self`, returning a new |SparseGenomeArray|
        and leaving `self` unchanged. `other` may be a scalar quantity or
        another |SparseGenomeArray|. In the latter case, multiplication is elementwise.
        
        Parameters
        ----------
        other : float, int, or |MutableAbstractGenomeArray|
        
        mode : str, choice of `'same'`, `'all'`, or `'truncate'`
            Only relevant if `other` is a |MutableAbstractGenomeArray|
        
            If '`same`' each set of corresponding chromosomes must
            have the same dimensions in both parent |SparseGenomeArrays|.
            In addition, both |SparseGenomeArrays| must have the same
            chromosomes, and the same strands.

            If '`all`', corresponding chromosomes in the parent
            |SparseGenomeArrays| will be resized to the dimensions of
            the larger chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; all chromosomes and strands will be
            included in output. 
                    
            If `'truncate'`,corresponding chromosomes in the parent
            |SparseGenomeArrays| will be resized to the dimensions of
            the smaller chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; chromosomes or strands not common
            to both parents will be ignored.                 
        
        Returns
        -------
        |SparseGenomeArray|
        """
        if isinstance(other, GenomeArray):
            new_array = SparseGenomeArray.like(self)
            chroms = set(self.keys()) & set(other.keys())
            strands = set(self.strands()) & set(other.strands())
            for chrom in chroms:
                for strand in strands:
                    new_array._chroms[chrom][strand] = self._chroms[chrom][strand].multiply(
                        other._chroms[chrom][strand]
                    )

            return new_array
        else:
            return self.apply_operation(other, operator.mul, mode=mode)

    def apply_operation(self, other, func, mode=None):
        """Apply a binary operator to a copy of `self` and to `other` elementwise.
        `other` may be a scalar quantity or another |GenomeArray|. In both cases,
        a new |SparseGenomeArray| is returned, and `self` is left unmodified.
        If :meth:`set_normalize` is set to `True`, it is disabled during the
        operation.
        
        Parameters
        ----------
        other : float, int, or |MutableAbstractGenomeArray|
            Second argument to `func`
        
        func : func
            Function to perform. This must take two arguments.
            a :py:class:`numpy.ndarray` (chromosome-strand) from the
            |SparseGenomeArray| will be supplied as the first argument, and the
            corresponding chromosome-strand from `other` as the second.
            This operation will be applied over all chromosomes and strands.
            
            If mode is set to `all`, and a chromosome or strand is not
            present in one of self or other, zero will be supplied
            as the missing argument to func. func must handle this
            gracefully.
        
        mode : str, choice of `'same'`, `'all'`, or `'truncate'`
            Only relevant if `other` is a |MutableAbstractGenomeArray|
        
            If '`same`' each set of corresponding chromosomes must
            have the same dimensions in both parent |SparseGenomeArrays|.
            In addition, both |SparseGenomeArrays| must have the same
            chromosomes, and the same strands.

            If '`all`', corresponding chromosomes in the parent
            |SparseGenomeArrays| will be resized to the dimensions of
            the larger chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; all chromosomes and strands will be
            included in output. 
                    
            If `'truncate'`,corresponding chromosomes in the parent
            |SparseGenomeArrays| will be resized to the dimensions of
            the smaller chromosome before the operation is applied.
            Parents are not required to have the same chromosomes
            or strands; chromosomes or strands not common
            to both parents will be ignored.   
        
        Returns
        -------
        |SparseGenomeArray|
            new |SparseGenomeArray| after the operation is applied
        """
        new_array = SparseGenomeArray.like(self)

        old_normalize = self._normalize
        if old_normalize == True:
            warn(
                "Temporarily turning off normalization during value set. It will be re-enabled automatically when complete.",
                DataWarning
            )
        self.set_normalize(False)

        if isinstance(other, GenomeArray):
            chroms = {}.fromkeys(set(self.keys()) | set(other.keys()), 10)
            strands = set(self.strands()) | set(other.strands())
            new_array = SparseGenomeArray(chroms, strands=strands)
            for chrom in chroms:
                if chrom in self.keys() and chrom in other.keys():
                    for strand in strands:
                        if strand in self.strands() and strand in other.strands():
                            sc = self._chroms[chrom][strand]
                            oc = other._chroms[chrom][strand]
                            new_array._chroms[chrom][strand] = func(sc, oc)
                        elif strand in self.strands():
                            new_array._chroms[chrom][strand] = func(
                                copy.deepcopy(self._chroms[chrom][strand]), 0
                            )
                        else:
                            new_array._chroms[chrom][strand] = func(
                                copy.deepcopy(other._chroms[chrom][strand]), 0
                            )
                elif chrom in self.keys():
                    for strand in strands:
                        new_array._chroms[chrom][strand] = func(
                            copy.deepcopy(self[chrom][strand]), 0
                        )
                else:
                    for strand in strands:
                        new_array._chroms[chrom][strand] = func(
                            copy.deepcopy(other[chrom][strand]), 0
                        )
        else:
            for chrom in self.keys():
                for strand in self.strands():
                    new_array._chroms[chrom][strand] = func(self._chroms[chrom][strand], other)

        self.set_normalize(old_normalize)
        return new_array

    def nonzero(self):
        """Return the indices of chromosomal positions with non-zero values
        at each chromosome/strand pair. Results are returned as a hierarchical
        dictionary mapping chromosome names, to a dictionary of strands,
        which in turn map to  to arrays of non-zero indices on that chromosome-strand.
        
        Returns
        -------
        dict
            `dict[chrom][strand]` = numpy.ndarray of indices
        """
        d_out = {}
        for key in self.keys():
            d_out[key] = {}
            for strand in self.strands():
                # need to sort because dok doens't guarantee sorted indices in nonzero()
                d_out[key][strand] = numpy.array(sorted(self._chroms[key][strand].nonzero()[1]))

        return d_out

    def _slicewrap(self, x):
        """Helper function to wrap coordinates for VariableStep/`bedGraph`_ export"""
        return (0, x)

    @staticmethod
    def like(other):
        """Return a |SparseGenomeArray| of same dimension as the input array
        
        Parameters
        ----------
        other : |GenomeArray| or |SparseGenomeArray|
        
        Returns
        -------
        |SparseGenomeArray|
            of same size as `other`
        """
        return SparseGenomeArray(other.lengths(), strands=other.strands())



#!/usr/bin/env python
"""Test cases for :py:mod:`plastid.util.io.binary`"""
from plastid.util.io.binary import BinaryParserFactory, find_null_bytes
import unittest
import re
from nose.plugins.attrib import attr
from io import BytesIO  #from plastid.util.services.mini2to3 import cStringIO


@attr(test="unit")
class TestBinaryParserFactory(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.factories = {"color": BinaryParserFactory("ColorFactory", "BBB", ["r", "g", "b"])}

        cls.results = {
            "color": [
                dict(zip(list("rgb"), X)) for X in [
                    (255, 0, 0),
                    (255, 255, 0),
                    (255, 255, 255),
                    (255, 0, 255),
                ]
            ]
        }

        cls.rawdata = { "color" : b"\xFF\x00\x00"+\
                                  b"\xFF\xFF\x00"+\
                                  b"\xFF\xFF\xFF"+\
                                  b"\xFF\x00\xFF"
                      }

        cls.size_dict = {
            "x": 0,
            "c": 1,
            "b": 1,
            "B": 1,
            "?": 1,
            "h": 2,
            "H": 2,
            "i": 4,
            "I": 4,
            "l": 4,
            "L": 4,
            "q": 8,
            "Q": 8,
            "f": 4,
            "d": 8,
            "s": 1,
            "p": 1,
        }

        cls.fmt_pat = re.compile(r"(?P<size>[0-9]+)*(?P<code>\w)")

    def call_helper(self, key):
        fh = BytesIO(self.rawdata[key])
        for result in self.results[key]:
            self.assertEqual(self.factories[key](fh), result)

        fh.close

    def test_call_color(self):
        self.call_helper("color")


@attr(test="unit")
def test_find_null_bytes():
    test = b"12345\x00123456789\x001234567890\x00"
    assert list(find_null_bytes(test)) == [5, 15, 26]



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.util.services.lists`"""
import unittest
import numpy
from plastid.util.services.lists import parse_list, flatten_nested_lists_to_generator, flatten_nested_lists_to_list
from nose.plugins.attrib import attr


@attr(test="unit")
class TestLists(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # list that will be entirely flattened
        cls.tree1 = [
            "a",
            "b",
            ["c", "d", "e"],
            [["f", "g", ["h"], ["i", ["j", "k"], "l"]], "m"],
            "n",
            ["o"],
            [["p"]],
            [["q"]],
            [["r", "s"]],
            ["t"],
        ]
        cls.list1 = list("abcdefghijklmnopqrst")

        # list containing tuples, which will not be flattened
        cls.tree2 = [
            "a",
            "b",
            ("c", "d", "e"),
            [["f", "g", ["h"], ["i", ["j", "k"], "l"]], "m"],
            "n",
            ["o"],
            [["p"]],
            [["q"]],
            (["r", "s"], ),
            ["t"]
        ]
        cls.list2 = [
            "a", 
            "b",
            ("c", "d", "e"),
            "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q",
            (["r", "s"], ),
            "t"
        ] 
 
    def test_parse_list(self):
        # homogeneous lists

        # positive ints
        self.assertEquals(parse_list(str(list(range(50)))), list(range(50)))

        # positive and negative ints
        self.assertEquals(parse_list(str(list(range(-50, 100)))), list(range(-50, 100)))

        # string
        a = ["asdfads", "basfdads", "asdfaj3r23fjd8c", "adsfj3r92h"]
        self.assertEquals(parse_list(str(a)), a)

        # positive and negative floats
        b = list(numpy.linspace(-5, 200, 43))
        self.assertEquals(parse_list(str(b)), b)

        # heterogeneous list
        b[10] = True
        b[20] = 3
        b[30] = False
        b[40] = ("this is a test")
        self.assertEquals(parse_list(str(b)), b)

    def test_flatten(self):
        for item1, item2 in zip(flatten_nested_lists_to_generator(self.tree1), self.list1):
            self.assertEquals(item1, item2)

        for item1, item2 in zip(flatten_nested_lists_to_generator(self.list1), self.list1):
            self.assertEquals(item1, item2)

        self.assertNotEquals(self.tree1, self.list1)

        for item1, item2 in zip(flatten_nested_lists_to_generator(self.tree2), self.list2):
            self.assertEquals(item1, item2)

        for item1, item2 in zip(flatten_nested_lists_to_generator(self.list2), self.list2):
            self.assertEquals(item1, item2)

        self.assertNotEquals(self.tree2, self.list2)

    def test_flatten_typesafe(self):
        self.assertRaises(AssertionError, next, flatten_nested_lists_to_generator(5))
        self.assertRaises(AssertionError, next, flatten_nested_lists_to_generator("some_string"))
        self.assertRaises(AssertionError, next, flatten_nested_lists_to_generator(("a", "b", "c")))

    def test_flatten_nested_lists(self):
        self.assertEquals(flatten_nested_lists_to_list(self.tree1), self.list1)
        self.assertEquals(flatten_nested_lists_to_list(self.list1), self.list1)
        self.assertNotEquals(self.tree1, self.list1)

        self.assertEquals(flatten_nested_lists_to_list(self.tree2), self.list2)
        self.assertEquals(flatten_nested_lists_to_list(self.list2), self.list2)
        self.assertNotEquals(self.tree2, self.list2)

    def test_flatten_nested_lists_typesafe(self):
        self.assertRaises(AssertionError, flatten_nested_lists_to_list, 5)
        self.assertRaises(AssertionError, flatten_nested_lists_to_list, "some_string")
        self.assertRaises(AssertionError, flatten_nested_lists_to_list, ("a", "b", "c"))



#!/usr/bin/env python
"""Base classes for functional tests, which compare script output to reference output

These are implemented as :py:obj:`nose` test generators, so that tests may
be evaluated at the level of individual output files, as opposed to suites
of entire test runs. Thus, any test modules based upon this module will require
:py:obj:`nose` to run.

To use these tools, define a dictionary describing the test information
and the tests that need to be run. Then create a dummy function that calls
:py:func:`execute_helper`::

    #!/usr/bin/env python
    '''This module contains a functional test'''

    from module.I.want.to.test import main
    from plastid.test.functional.base import execute_helper
    from plastid.util.services.decorators import suppress_stderr
    
    my_test_info = { 
        "test_method"    : suppress_stderr(main), # hide output from screen
        "module_name"    : "module.I.want.to.test",
        "ref_file_path"  : resource_filename("plastid","test/data/command_line"),
        "temp_file_path" : tempfile.mkdtemp(prefix="test_folder"),    
     }
     
    my_tests = [ ... ]
    
    def do_test():
        for x in execute_helper(my_test_info,my_tests):
            yield x
                
If run from the command-line, this script will generate modules for functional
tests of command-line scripts based upon a template resembling that above.
These modules may then be edited to flesh out the tests as necessary.
"""
import argparse
import sys
import shlex
import shutil
import inspect
import os
import nose

from nose.tools import assert_true
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import get_short_name, NullWriter
from plastid.bin.test_table_equality import main
from plastid.util.services.decorators import skip_if_abstract, catch_stderr
from pkg_resources import resource_filename, cleanup_resources

printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))
table_test = catch_stderr()(main)

#===============================================================================
# INDEX: test generator functions
#===============================================================================


def check_one_file(test_info, ref_file, output_file, eq_args):
    """Check a single output file produced by a test run against a reference file 
    
    Parameters
    ----------
    test_info : dict
        Dictionary containing miscellaneous test information
    
    ref_file : str
        Name of reference file, excluding path, against which test output will be checked
    
    output_file : str
        Name of output file, excluding path
    
    eq_args : str
        Equality arguments determining how the reference and output file
        should be compared by :py:func:`table_test`
    """
    retval, failures = table_test([ref_file, output_file] + shlex.split(eq_args), verbose=True)
    msg = "\n    ".join(failures)
    err_message = "Unequal output in files %s vs %s, module %s:\n    %s" % (
        ref_file, output_file, test_info["module_name"], msg
    )
    nose.tools.assert_equal(retval, 0, msg=err_message)


def build_test_lists(test_info, test):
    """Build a list of tests to be performed after one execution of a command-line script.
    
    Parameters
    ----------
    test_info : dict
        Dictionary containing miscellaneous test information
    
    test : list
        list of 3-element lists, each consisting of:
        
            1.  A reference filename against which test output will be compared
            
            2.  A filename of test output
            
            3.  A string of equality arguments determining how the reference
               and output file should be compared by :py:func:`table_test`
    
    Returns
    -------
    list
        list of:
        
        one 1-tuple describing arguments to past to the test method to run.
        This is always the first item in the list.
        
        The remaining items are one or more 5-tuples describing all the tests
        that must be run to cover each output file produced by the test method run.
        These contain:
        
            1.  the test function :py:func:`check_one_file`
            
            2.  the dictionary ``test_info``, describing file paths and other data
            
            3.  The file name of the reference output file for the test, 
                which the test output should match.
            
            4.  The corresponding test output filename for that test
            
            5.  Equality arguments for that test, to pass to :py:func:`table_test`
    """
    argstr, ref_files, test_files, eqarg_groups = test
    test_list = []
    test_list.append([argstr])
    for ref_file, output_file, eq_args in zip(ref_files, test_files, eqarg_groups):
        test_list.append((check_one_file, test_info, ref_file, output_file, eq_args))

    return test_list


def execute_helper(test_info, tests):
    """Execute functional tests and evaluate all of their output files.
    This is a test generator function for use with :py:obj:`nose`. To useit,
    import it into a test module, and create a test function as follows::
    
        from plastid.test.functional.base import execute_helper
        my_test_info = { ... } # define this
        my_tests = [ ... ] # define tests. see Parameters section, below
        
        def do_test():
            for x in execute_helper(my_test_info,my_tests):
                yield x
    
    
    Parameters
    ----------
    test_info : dict
        Dictionary containing miscellaneous test information
    
    tests : list
        list of 4 element lists, each consisting of:
            
            1. Command-line arguments to pass to ``test_info["test_method"]``
            
            2. A list of reference files, which output files created by the test should match
             
            3. A list of output files created by the test
            
            4. A list of equality arguments (as strings) determining how each
               pair of reference and output files should be compared
               by :py:func:`table_test`
        
    Yields
    ------
    tuple
        5-tuples, collectively describing all the tests that must be run to cover
        every output file produced by every test run. Each tuple contains:
        
            1.  the test function :py:func:`check_one_file`
            
            2.  the dictionary ``test_info``, describing file paths and other data
            
            3.  The file name of the reference output file for the test, 
                which the test output should match.
            
            4.  The corresponding test output filename for that test
            
            5.  Equality arguments for that test, to pass to :py:func:`table_test`  
    """
    test_list = []
    for test in tests:
        _, ref_files, test_files, eqarg_groups = test
        assert len(eqarg_groups) == len(test_files) == len(ref_files)
        test_list.extend(build_test_lists(test_info, test))

    for test_item in test_list:
        # if len 1, item is an execute block, which will generate files to be tested next
        if len(test_item) == 1:
            # run test, but do it in try-catch so that if there is an execution error
            # subsequent tests will still run
            try:
                print("Executing test %s" % test_item[0])
                test_info["test_method"](shlex.split(test_item[0]))
            except Exception as e:
                # report that there was an error
                yield check_item_error, test_item, e
        # otherwise, item is a test
        else:
            yield test_item

    if test_info["temp_file_path"] != "":
        shutil.rmtree(test_info["temp_file_path"])


def check_item_error(argstr, e):
    """Placeholder function used to raise exceptions when specific test executions fail
    
    Parameters
    ----------
    argstr : str
        String of arguments that were passed to test

    e : Exception
        Exception that was raised
    """
    assert_true(False, "Error executing test %s\n%s" % (argstr, e))


#===============================================================================
# INDEX: command-line program
#===============================================================================


def create_test_suite(module_names, base_path, overwrite=False, printer=NullWriter()):
    """Creates a :py:class:`unittest.TestCase` for the ``main()`` method of
    a command-line python module, and write these to modules
    
    Parameters
    ----------
    module_names : list
        List of module names, as strings, to create :py:class:`~unittest.TestCase` s for
    
    base_path : str
        Path where test modules will be written
    
    overwrite : bool, optional
        If True, existing files will be overwritten (Default: False)
    
    printer : file-like, optional
        Logger to which output will be written
    """
    for module_name in module_names:
        printer.write(module_name)
        short_name = get_short_name(module_name, separator="\.", terminator="")
        output_str = TEST_TEMPLATE.replace("${MODULE}", module_name)
        output_str = output_str.replace("${CAP_SHORT_NAME}", short_name.capitalize())
        output_str = output_str.replace("${SHORT_NAME}", short_name)
        output_fn = os.path.join(base_path, "test_%s.py" % short_name)
        if overwrite or not os.path.exists(output_fn):
            printer.write("Writing %s" % output_fn)
            with open(output_fn, "w") as fout:
                fout.write(output_str)
                fout.close()


def main(argv=sys.argv[1:]):
    """Command-line program to generate functional test suites for command-line
    scripts from a template based upon the utilities in this module.
    
    For each test, output for given parameter sets is compared to reference output.
    
    Files will not be overwritten of they are already present.
    
    Parameters
    ----------
    argv : list
        list of command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "bin_folder", type=str, help="Location of command-line modules to write tests for"
    )
    parser.add_argument(
        "output_folder", type=str, help="Location of folder in which to write test modules"
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="If supplied, overwrite existing files"
    )

    args = parser.parse_args(argv)
    modules   = [("plastid.bin.%s" % X).replace(".py","") \
                 for X in filter(lambda x: x.endswith(".py") and "__init__" not in x,
                                 os.listdir(args.bin_folder))]
    create_test_suite(modules, args.output_folder, overwrite=args.overwrite, printer=printer)


#===============================================================================
# INDEX : test template
#===============================================================================

TEST_TEMPLATE = '''#!/usr/bin/env python
"""Test suite for :py:mod:`${MODULE}`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename
from plastid.util.services.decorators import catch_stderr
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import REF_FILES, \\
                                       RPATH, \\
                                       REF_FILES, \\
                                       COUNT_OPTIONS, \\
                                       ANNOTATION_OPTIONS, \\
                                       MASK_OPTIONS
from ${MODULE} import main

test_info = {
    "test_method"    : catch_stderr()(main),
    "module_name"    : "${MODULE}",
    "ref_file_path"  : resource_filename("plastid","test/data/command_line"),
    "temp_file_path" : tempfile.mkdtemp(prefix="${SHORT_NAME}"),
}

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
${SHORT_NAME}_tests = [
    ("",[],[],[]),
    ("",[],[],[]),
    ("",[],[],[]),
    ("",[],[],[]),
    ("",[],[],[]),
]
"""Functional tests of :py:mod:`${MODULE}`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod`${MODULE}`"""
    for x in execute_helper(test_info,${SHORT_NAME}_tests):
        yield x
'''
"""Template for functional tests using test generators from :py:obj:`nose`"""

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Compare splice junctions discovered in a dataset to those in an annotation
of known splice junctions, amending misplaced junctions, and identifying 
junctions that fall within repetitive areas of the genome.

Known splice junctions can be misidentified as novel or non-canonical junctions
when intronic sequence immediately downstream of the fiveprime splice site
exactly matches the exonic sequence immediately downstream of the threeprime
splice site. In fact, the junction point could appear anywhere in this
locally-repeated region with equal support from sequencing data. For example,
suppose we have a splice junction as follows::

        
                Exon 1 [0,6)            Intron                                  Exon 2 [16,24)
                ---------------------   --------------------------------------  ------------------------------
    Sequence    G   C   T   C   T   A   C   T   A   G   N   N   N   C   T   A   C   T   A   G   A   T   G   G
    Position    0   1   2   3   4   5   6   7   8   9   10  11  12  13  14  15  16  17  18  19  20  21  22  23
    Repeated                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    
In this case, the splice junction could be moved 3 bases to the left, or
four bases to the right, without losing consistency with the sequence of
any cDNA or read alignment covering the junction.
   
To identify this and other causes of false positive splice junction calls,
the following operations are performed on each query junction:

    1.  If a :term:`mask file` from |crossmap|
        is provided, junctions in which one or more of the 5' and 3' splice
        sites appear in a repetitive region of the genome are flagged as
        non-informative and written to a separate file. 

    2.  For remaining splice junctions, the extent of locally repeated nucleotide
        sequence, if any, surrounding the query junction's splice donor and
        acceptor sites, are determined in both the 5' and 3' directions.
        
        This is the maximum window (*equal-support region*) across which the
        actual splice junction could be moved without reducing sequence support.
    
    3.  If there is one or more known splice junctions in this region, the
        query junction is assumed to match these, and the known junctions are
        reported rather than the query. 
    
    4.  If (3) is not satisfied, and the query junction is a canonical splice
        junction, it is reported as is.

    5.  If (3) is not satisfied, and the query junction represents a non-canonical
        splice junction, the program determines if one or more canonical splice
        junctions is present in the equal-support region. If so, these canonical
        splice junction are reported rather than the query junction.
        
    6.  If (5) is not satisfied, the non-canonical query junction is reported as-is.

Output files
------------
The following files are written, where `OUTBASE` is a string supplied by the
user. Scores of splice junctions, if present in the input, are ignored.
Each record in each `BED`_ file represents a single exon-exon junction, rather
than a transcript:

    OUTBASE_repetitive.bed
        Splice junctions in which one or more of the splice sites lands
        in a repetitive/degenerate region of the genome, which gives rise to
        mapping ambiguities (step 1 above)
    
    OUTBASE_shifted_known.bed
        The result of shifting query splice junctions to known splice junctions
        with equal sequence support (step 3 above)
    
    OUTBASE_shifted_canonical.bed
        The result of shifting non-canonical query splice junctions to canonical
        splice junctions with equal sequence support (step 5 above)
    
    OUTBASE_untouched.bed
        Query junctions reported without changes (steps 4 and 6 above)

where `OUTBASE` is given by the user.

"""
__date__ = "Aug 23, 2011"
__author__ = "joshua"

import sys
import argparse
from plastid.util.io.openers import opener, argsopener, get_short_name
from plastid.util.io.filters import CommentReader, NameDateWriter
from plastid.genomics.splicing import get_junction_tuple
from plastid.genomics.roitools import GenomicSegment, SegmentChain
from plastid.genomics.genome_hash import GenomeHash
from plastid.readers.bed import BED_Reader
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.scriptlib.argparsers import (MaskParser, SequenceParser, BaseParser)
import inspect
import warnings
warnings.simplefilter("once")

printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

#===============================================================================
# INDEX : helper functions that classify/manipulate splice junctions
#===============================================================================


def find_match_range(seg, genome, maxslide):
    """Find maximum distance over which a splice junction can be moved up-
    or down-stream without reducing sequencing support for that junction.
    
    In other words, find locally repeated sequences surrounding exon-intron
    or intron-exon boundaries that can cause splice junction mapping to be
    ambiguous, due to identical and repeated sequence.

    Parameters
    ----------
    seg : |SegmentChain|
         A two-exon fragment representing a query splice junction
         
    genome : dict
        dict mapping chromosome names to :py:class:`Bio.SeqRecord.SeqRecord` s
        
    maxslide : int
        Maximum number of nucleotides from the boundary over which to check
        for extent of repeated sequence
    
    Returns
    -------
    minus_range : int
        Maximum number of nucleotides splice junction point could be moved 
        to the left without reducing sequence support for the junction
        
    plus_range : int
        Maximum number of nucleotides splice junction point could be moved 
        to the right without reducing sequence support for the junction
    """
    iv1, iv2 = seg[0], seg[1]
    chrom = seg.chrom

    plus_range = 0
    minus_range = 0
    check_plus = True
    check_minus = True

    # Discovery of match range. Suppose we have a splice junction as follows:
    #
    #        Exon 1 [0,6)            Intron                                  Exon 2 [16,24)
    #        ---------------------   --------------------------------------  ------------------------------
    # Seq    G   C   T   C   T   A   C   T   A   G   N   N   N   C   T   A   C   T   A   G   A   T   G   G
    # Pos    0   1   2   3   4   5   6   7   8   9   10  11  12  13  14  15  16  17  18  19  20  21  22  23
    #
    #
    # Exon 1 start: 0
    # Exon 1 end:   6
    # Exon 2 start: 16
    # Exon 2 end:   24
    #
    # Match range: -3, 4
    #
    # Then seq[ex1.end    ] == seq[6] == C == seq[16] == seq[ex2.start    ]
    # ---------------------------------------------------------------------
    # Then seq[ex1.end - 3] == seq[3] == C == seq[13] == seq[ex2.start - 3]
    # Then seq[ex1.end - 2] == seq[4] == T == seq[14] == seq[ex2.start - 2]
    # Then seq[ex1.end - 1] == seq[5] == A == seq[15] == seq[ex2.start - 1]
    # Then seq[ex1.end + 0] == seq[6] == C == seq[16] == seq[ex2.start + 0]
    # Then seq[ex1.end + 1] == seq[7] == T == seq[17] == seq[ex2.start + 1]
    # Then seq[ex1.end + 2] == seq[8] == A == seq[18] == seq[ex2.start + 2]
    # Then seq[ex1.end + 3] == seq[9] == G == seq[19] == seq[ex2.start + 3]
    #
    # We must therefore search for known, repetitive, or canonical splice junctions
    # by iterating our offset ``i`` over [-3,3] = ``range(minus_range, plus_range+1)``

    for i in range(maxslide + 1):
        if check_plus is True and\
        genome[chrom][iv1.end + i] == genome[chrom][iv2.start + i]:
            plus_range = i + 1  #+=1
        else:
            check_plus = False

        if check_minus is True and i > 0:
            if genome[chrom][iv1.end - i] == genome[chrom][iv2.start - i]:
                minus_range = -i  #-=1
            else:
                check_minus = False

    return minus_range, plus_range


def find_known_in_range(query_junc, minus_range, plus_range, knownjunctions):
    """Find any known splice junctions within in `minus_range...plus_range`
    of `query_junc`
    
    To be classified as within the range, the boundaries of a known
    junction must be:
    
      1. within `minus_range...plus_range` of the boundaries
         of the the discovered junction.
            
      2. separated by a nucleotide distance equal to the distance 
         separating the junction in `query_junc`.
        
      3. on the same chromosome and strand.
    
    
    Parameters
    ----------
    query_junc : |SegmentChain|
         A two-exon fragment representing a query splice junction
         
    minus_range : int <= 0
        Maximum number of nucleotides splice junction could be moved 
        to the left without reducing sequence support for the junction
        see :py:func:`find_match_range`
        
    plus_range : int >= 0
        Maximum number of nucleotides splice junction could be moved 
        to the right without reducing sequence support for the junction
        see :py:func:`find_match_range`
    
    knownjunctions : list of |SegmentChains|
        known splice junctions
        
    Returns
    -------
    list
        List of |SegmentChains| representing known splice junctions in
        `minus_range...plus_range` of `query_junc`
    """
    exact_match = None
    ltmp = []
    iv1, iv2 = query_junc[0], query_junc[1]
    iv1end = iv1.end
    iv2start = iv2.start
    qstrand = query_junc.strand
    qchrom = query_junc.chrom

    for ivc in knownjunctions:
        if qchrom != ivc.chrom:
            continue
        elif qstrand != ivc.strand:
            continue
        elif query_junc == ivc:
            exact_match = query_junc
            break
        kiv1, kiv2 = ivc[0], ivc[1]
        kend = kiv1.end
        kstart = kiv2.start
        if  kend   in set(range(iv1end+minus_range,  iv1end  +plus_range+1))\
        and kstart in set(range(iv2start+minus_range,iv2start+plus_range+1)):
            if kstart - kend == iv2start - iv1end:
                ltmp.append(ivc)
    return [exact_match] if exact_match is not None else ltmp


def find_canonicals_in_range(query_junc, minus_range, plus_range, genome, canonicals):
    """Find any canonical splice junctions within in `minus_range...plus_range`
    of `query_junc`
    
    To be classified as within the range, the boundaries of the canonical
    junction must be:
    
      1. within `minus_range...plus_range` of the boundaries
         of the the discovered junction.
            
      2. separated by a nucleotide distance equal to the distance 
         separating the junction in `query_junc`.
        
      3. On the same chromosome and strand.
        
    
    Parameters
    ----------
    query_junc : |SegmentChain|
         A two-exon fragment representing a query splice junction
         
    minus_range : int <= 0
        Maximum number of nucleotides splice junction could be moved 
        to the left without reducing sequence support for the junction
        see :py:func:`find_match_range`
        
    plus_range : int >= 0
        Maximum number of nucleotides splice junction could be moved 
        to the right without reducing sequence support for the junction
        see :py:func:`find_match_range`
        
    genome : dict
        dict mapping chromosome names to :py:class:`Bio.SeqRecord.SeqRecord` s
        
    canonicals : list
        dinucleotide sequences to consider as canonical splice sites,
        as a list of tuples. e.g. `[("GT","AG"), ("GC","AG")]`

        
    Returns
    -------
    list
        List of |SegmentChains| representing canonical splice junctions in
        `minus_range...plus_range` of `query_junc`
    """
    ltmp = []
    chrom = query_junc.chrom
    strand = query_junc.strand
    iv1, iv2 = query_junc[0], query_junc[1]

    iv1start, iv1end = iv1.start, iv1.end
    iv2start, iv2end = iv2.start, iv2.end
    for i in range(minus_range, plus_range + 1):
        for pair in canonicals:
            if str(genome[chrom][iv1end + i:iv1end + i + 2].seq) == pair[0]\
            and str(genome[chrom][iv2start - 2 + i:iv2start + i].seq) == pair[1]:
                new_iv1 = GenomicSegment(chrom, iv1start, iv1end + i, strand)
                new_iv2 = GenomicSegment(chrom, iv2start + i, iv2end, strand)
                ltmp.append(SegmentChain(new_iv1, new_iv2))

    return ltmp


def covered_by_repetitive(query_junc, minus_range, plus_range, cross_hash):
    """Determine whether one or both ends of a splice site overlap with
    a repetitive area of the genome.
    
    Parameters
    ----------
    query_junc : |SegmentChain|
         A two-exon fragment representing a query splice junction
    
    minus_range : int <= 0
        Maximum number of nucleotides splice junction could be moved 
        to the left without reducing sequence support for the junction
        see :py:func:`find_match_range`
        
    plus_range : int >= 0
        Maximum number of nucleotides splice junction could be moved 
        to the right without reducing sequence support for the junction
        see :py:func:`find_match_range`
    
    cross_hash : |GenomeHash|
        |GenomeHash| of 1-length features denoting repetitive regions of the genome
        
    
    Returns
    -------
    bool
        `True` if any of the genomic positions within `minus_range...plus_range`
        of the 5' or 3' splice sites of `query_junc` overlap a repetitive
        region of the genome as annotated by ``cross_hash``.
        Otherwise, `False`
    """
    chrom = query_junc.spanning_segment.chrom
    strand = query_junc.spanning_segment.strand
    qend = query_junc[0].end
    qstart = query_junc[1].start
    fiveprime_splice_area = GenomicSegment(chrom, qend + minus_range, qend + plus_range + 1, strand)
    threeprime_splice_area = GenomicSegment(
        chrom, qstart + minus_range, qstart + plus_range + 1, strand
    )
    support_region = SegmentChain(fiveprime_splice_area, threeprime_splice_area)
    return len(cross_hash.get_overlapping_features(support_region)) > 0


#===============================================================================
# INDEX : program body
#===============================================================================


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    sp = SequenceParser()
    mp = MaskParser()
    bp = BaseParser()

    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[bp.get_parser(), sp.get_parser(),
                 mp.get_parser()],
    )
    parser.add_argument(
        "--maxslide",
        type=int,
        default=10,
        help="Maximum number of nt to search 5\' and 3\' of intron" + " boundaries (Default: 10)"
    )
    parser.add_argument(
        "--ref",
        type=str,
        metavar="ref.bed",
        default=None,
        help="Reference file describing known splice junctions"
    )
    parser.add_argument(
        "--slide_canonical",
        action="store_true",
        default=False,
        help="Slide junctions to canonical junctions if present within equal support region"
    )
    parser.add_argument(
        "infile", type=str, metavar="input.bed", help="BED file describing discovered junctions"
    )
    parser.add_argument("outbase", type=str, help="Basename for output files")
    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    printer.write("Opening genome from %s..." % args.sequence_file)
    genome = sp.get_seqdict_from_args(args)

    # load crossmap
    cross_hash = mp.get_genome_hash_from_args(args)

    # load ref junctions
    if args.ref is not None:
        printer.write("Loading reference junctions from %s" % args.ref)
        known_hash = GenomeHash(list(BED_Reader(open(args.ref))), do_copy=False)
    else:
        known_hash = GenomeHash()

    # set up variables
    canonicals_plus = [("GT", "AG"), ("GC", "AG")]

    canonicals_minus = [("CT", "AC"), ("CT", "GC")]

    known_in_range = 0
    canonical_in_range = 0
    repetitive = 0
    untouched = 0
    c = 0

    seen_already = []

    outfiles = {
        "repetitive": "%s_repetitive.bed" % args.outbase,
        "known"     : "%s_shifted_known.bed" % args.outbase,
        "canonical" : "%s_shifted_canonical.bed" % args.outbase,
        "untouched" : "%s_untouched.bed" % args.outbase,
    } # yapf: disable
    outfiles = {K: argsopener(V, args, "w") for K, V in outfiles.items()}

    # process data
    printer.write("Opening junctions from %s..." % args.infile)
    for ivc in BED_Reader(CommentReader(opener(args.infile))):
        processed = False
        tup = None

        if c % 1000 == 0 and c > 0:
            printer.write("Processed: %s\tknown: %s\tshifted to canonical: %s\trepetitive: %s\tuntouched: %s" % \
                    (c, known_in_range, canonical_in_range, repetitive, untouched))

        assert len(ivc) == 2

        strand = ivc.strand
        minus_range, plus_range = find_match_range(ivc, genome, args.maxslide)

        # see if either end of splice junction +- match_range lands in repetitive areas of genome
        if covered_by_repetitive(ivc, minus_range, plus_range, cross_hash):
            repetitive += 1
            outfiles["repetitive"].write(ivc.as_bed())
            processed = True

        # see if one or more known junctions in range
        if processed == False and args.ref is not None:
            # find_known_in_range(query_ivc,minus_range,plus_range,knownjunctions)
            known_juncs = find_known_in_range(
                ivc, minus_range, plus_range, known_hash.get_nearby_features(ivc)
            )
            if len(known_juncs) > 0:
                known_in_range += 1
                for my_known in known_juncs:
                    tup = get_junction_tuple(my_known)
                    if tup not in seen_already:
                        outfiles["known"].write(my_known.as_bed())
                        seen_already.append(tup)

                processed = True

        # see if one or more canonical junctions in range
        if processed == False and args.slide_canonical == True:
            canonicals = canonicals_plus if strand == "+" else canonicals_minus
            #find_canonicals_in_range(query_ivc,minus_range,plus_range,genome,canonicals)
            canonical_juncs = find_canonicals_in_range(
                ivc, minus_range, plus_range, genome, canonicals
            )
            if len(canonical_juncs) > 0:
                canonical_in_range += 1
                for can in canonical_juncs:
                    tup = get_junction_tuple(can)
                    if tup not in seen_already:
                        outfiles["canonical"].write(can.as_bed())
                        seen_already.append(tup)

                processed = True

        if processed == False:
            outfiles["untouched"].write(ivc.as_bed())
            untouched += 1

        c += 1

    # save output
    printer.write("Totals: %s\tknown: %s\tshifted to canonical: %s\trepetitive: %s\tuntouched: %s" % \
            (c, known_in_range, canonical_in_range, repetitive, untouched))

    for v in outfiles.values():
        v.close()

    printer.write("Done.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Constants, functions, and classes used by multiple readers in this subpackage


Functions & classes
-------------------
:func:`get_identical_attributes`
    Return a dictionary of all key-value pairs that are common to all `attr`
    dictionaries in a set of |SegmentChains|
    
|AssembledFeatureReader|
    Base class for readers that assemble high-level features (e.g. gapped
    alignments or transcripts) from one or more sub-features in an annotation file

"""
import itertools
import pysam

try:
    from pysam.ctabix import tabix_generic_iterator, tabix_file_iterator
except ImportError:
    from pysam.libctabix import tabix_generic_iterator, tabix_file_iterator

from plastid.util.io.filters import AbstractReader
from plastid.util.io.openers import NullWriter, multiopen
from plastid.genomics.roitools import (
    GenomicSegment,
    SegmentChain,
    Transcript,
    add_three_for_stop_codon,
)
from abc import abstractmethod

#===============================================================================
# INDEX: helper functions
#===============================================================================


def get_identical_attributes(features, exclude=None):
    """Return a dictionary of all key-value pairs that are identical for all |SegmentChains| in `features`
    
    Parameters
    ----------
    features : list 
        list of |SegmentChains|

    exclude : set
        attributes to exclude from identity criteria
    
    Returns
    -------
    dict
        Dictionary of all key-value pairs that have identical values in all the
        `attr` dictionaries of all the features in `features`
    """
    exclude = [] if exclude is None else exclude
    common_keys = set(features[0].attr.keys())
    for feature in features:
        common_keys &= set(feature.attr.keys())

    common_keys -= set(exclude)

    dtmp = { K : features[0].attr[K] for K in common_keys \
                 if all([X.attr[K] == features[0].attr[K] for X in features]) == True }
    return dtmp


#===============================================================================
# INDEX: classes
#===============================================================================


class AssembledFeatureReader(AbstractReader):
    """
    AssembledFeatureReader(*streams, return_type=SegmentChain, add_three_for_stop=False, tabix=False, printer=None, **kwargs)
    
    Abstract base class for readers that yield complex or discontinuous features
    such as transcripts or gapped alignments.
    
    For memory efficiency, all readers function as iterators. Readers built
    by subclassing |AssembledFeatureReader| are responsible for:
    
      - choosing when to yield assembled features
      
      - deciding how many subfeatures to hold in memory
      
      - overloading :meth:`~AssembledFeatureReader._assemble`
      

    Parameters
    ----------
    *streams : file-like
        One or more open filehandles of input data.
    
    return_type : |SegmentChain| or subclass, optional
        Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

    add_three_for_stop : bool, optional
        Some annotation files exclude the stop codon from CDS annotations. If set to
        `True`, three nucleotides will be added to the threeprime end of each
        CDS annotation, **UNLESS** the annotated transcript contains explicit stop_codon 
        feature. (Default: `False`)
                    
    printer : file-like, optional
        Logger implementing a ``write()`` method. Default: |NullWriter|
    
    tabix : boolean, optional
        `streams` are `tabix`_-compressed (Default: `False`)

    **kwargs
        Other keyword arguments used by specific parsers
      
    
    Attributes
    ----------
    streams : file-like
        Input streams, usually constructed from or more open filehandles
    
    metadata : dict
        Various attributes gleaned from the stream, if any

    counter : int
        Cumulative line number counter over all streams

    printer : file-like, optional
        Logger implementing a ``write()`` method.

    return_type : class
        The type of object assembled by the reader. Typically an |SegmentChain|
        or a subclass thereof.
    
    rejected : list
        A list of transcript IDs that failed to assemble properly
    """

    def __init__(self, *streams, **kwargs):
        """
        AssembledFeatureReader(*streams, return_type=SegmentChain, add_three_for_stop=False, printer=None, tabix=False, **kwargs)
        
        Parameters
        ----------
        streams : file-like
            One or more fileneames or open filehandles of input data.
        
        return_type : |SegmentChain| or subclass, optional
            Type of feature to return from assembled subfeatures (Default: |SegmentChain|)

        add_three_for_stop : bool, optional
            Some annotation files exclude the stop codon from CDS annotations. If set to
            `True`, three nucleotides will be added to the threeprime end of each
            CDS annotation, **UNLESS** the annotated transcript contains explicit stop_codon 
            feature. (Default: `False`)
                        
        printer : file-like, optional
            Logger implementing a ``write()`` method. Default: |NullWriter|
        
        tabix : boolean, optional
            `streams` point to `tabix`_-compressed files or are open
            :class:`~pysam.ctabix.tabix_file_iterator` (Default: `False`)

        **kwargs
            Other keyword arguments used by specific parsers
        """
        streams = multiopen(streams, fn=open, kwargs=dict(mode="rb"))

        if kwargs.get("tabix", False) == True:
            self.stream = itertools.chain.from_iterable((_tabix_iteradaptor(X) for X in streams))
        else:
            self.stream = itertools.chain.from_iterable(streams)

        self.counter = 0

        self.printer = kwargs.get("printer", NullWriter())

        self.return_type = kwargs.get("return_type", SegmentChain)
        add_three_for_stop = kwargs.get("add_three_for_stop", False)
        self._finalize = add_three_for_stop_codon if add_three_for_stop == True else lambda x: x

        self.metadata = {}
        self.rejected = []

    @abstractmethod
    def _assemble(self, data):
        """Assemble features from data. This must be implemented in subclass.
        
        Returns
        -------
        |SegmentChain| or subclass
            Next feature assembled from `self.streams`, type specified by `self.return_type`
        """

    def filter(self, data):
        """Return next assembled feature from `self.stream`
        
        Returns
        -------
        |SegmentChain| or subclass
            Next feature assembled from `self.streams`, type specified by `self.return_type`
        """
        return self._finalize(self._assemble(data))


def _tabix_iteradaptor(stream):
    """Open `stream` as an iterator over a `tabix`_ file, returning raw strings from tabix data.
    
    Parameters
    ----------
    streams : open file-like, :class:`pysam.ctabix.tabix_file_iterator`
    
    Returns
    -------
    generator
        Generator of tab-delimited string records in `tabix`_ file
    """
    if not isinstance(stream, (tabix_generic_iterator, tabix_file_iterator)):
        stream = tabix_file_iterator(stream, pysam.asTuple())

    return (str(X) for X in stream)



""":term:`Mapping functions` used by |BAMGenomeArray|

.. contents::
   :local:

Summary
-------
Numerous sequencing assays encode interesting biology in various properties of
read alignments. For example:

  - Ribosome profiling encodes the positions of ribosomal P-sites as a function
    of both read length and alignment coordinates

  - Bisulfite sequencing encodes methylation status as C-to-T transitions
    within read alignments

  - DMS-Seq and Pseudouridine profiling encode unstructured regions of RNA or
    sites of pseudouridine modification, respectively, in the 5' termini of
    their read aligments

Extended summary
----------------

plastid uses configurable mapping functions to decode the biology of interest
from read alignments.

This design enables plastid's tools to operate on sequencing data from
virtually any NGS assay, provided the appropriate  mapping function and
parameters. Mapping rules are described in the following section:

This module contains Cython implementations of mapping functions for BAM files.
At present, the following are included:


*Fiveprime end mapping*
   Each read alignment is mapped to its 5' end, or at a fixed offset (in
   nucleotides) from its 5' end

*Variable fiveprime end mapping*
   Each read alignment is mapped at a fixed distance from its 5' end, where
   the distance is determined by the length of the read alignment.

   This is used for :term:`ribosome profiling` of yeast (:cite:`Ingolia2009`)
   and mammalian cells (:cite:`Ingolia2011`).

*Stratified variable fiveprime end mapping*
   This multidimensional mapping rule behaves as variable fiveprime end mapping,
   and additionally returns arrays that are stratified by read length.

   For each region of interest, a 2D array is returned, in which each row
   represents a given read length, and each column a nucleotide position in the
   region of interest. In other words, summing over the columns produces the
   same array that would be given by variable fiveprime end mapping.

*Threeprime end mapping*
   Each read alignment is mapped to its 3' end, or at a fixed offset (in
   nucleotides) from its 3' end.

*Entire* or *Center-weighted mapping*
    Zero or more positions are trimmed from each end of the read alignment, and
    the remaining `N` positions in the alignment are incremented by `1/N` read
    counts (so that each read is still counted once, when integrated over its
    mapped length).

   This is also used for ribosome profiling of E. coli and D. melanogaster, and
   RNA-seq.



Examples
--------

Mapping functions are rarely called directly. Instead, they are invoked via
:meth:`plastid.genomics.genome_array.BAMGenomeArray.set_mapping`:

 .. code-block:: python

    # open BAM file
    >>> ga = BAMGenomeArray(["some_file.bam"])

    # set mapping 15 bases from 3' end of read
    >>> ga.set_mapping(ThreePrimeMapFactory(offset=15))

    # set mapping to 5' end of read
    >>> ga.set_mapping(FivePrimeMapFactory())

    # cut 3 nt off each end of the read, and map fractionally over the rest
    >>> ga.set_mapping(CenterMapFactory(nibble=12))

    # access reads in some region
    >>> ga[GenomicSegment("chrI", 10000, 100000, "+")]
    # output is numpy array

Other mapping functions are similarly invoked. See their docstrings for further
details.


Implementation
--------------

Mapping functions must be callables, and can therefore be implemented as functions
or classes. In order to give them configurable parameters, one could create
function factories or use :func:`functools.partial` to supply the extra
parameters. Here, we created classes that produce callable instances that mimic
functions. We did this because early on we ran into trouble with limitations of
:mod:`pickle` and :mod:`multiprocessing`, which can only capture functions and
classes defined in the top-level scope. This limitation is no longer relevant
to our purposes but is preserved here in history.

Generally, the callable must accept two arguments: a list of read alignments
(as :class:`pysam.AlignedSegment`) and a |GenomicSegment| representing a query
interval. The callable must then return two values: a list of read alignments,
and a :class:`numpy.ndarray` of values.

Generally, all reads are mapped within the callable. Only those that contribute
data to coordinates contained within the query interval are returned.

The :class:`numpy.ndarray` of values may be multidimensional. The only
stipulation is that the last dimension represent the positions in the query
interval. For a 2D array, the positions would thus be columns. See
|StratifiedVariableFivePrimeMapFactory| below for an example.


See also
--------
Module documentation for :mod:`~plastid.genomics.genome_array`
    For a detailed discussion of :term:`mapping functions <mapping rule>`.
"""

import numpy as np
cimport numpy as np
cimport cython

IF PYSAM10:
    from pysam.libcalignmentfile cimport AlignedSegment
ELSE:
    from pysam.calignmentfile cimport AlignedSegment

from plastid.genomics.c_common cimport forward_strand, reverse_strand, unstranded
from plastid.genomics.roitools cimport GenomicSegment
from plastid.util.services.exceptions import DataWarning, warn, warn_onceperfamily
from plastid.util.io.filters import CommentReader
from plastid.util.scriptlib.argparsers import _parse_variable_offset_file

DEF _BAD_OFFSET = -1
INT    = np.int
FLOAT  = np.float
DOUBLE = np.double
LONG   = np.long

ctypedef np.int_t    INT_t
ctypedef np.float_t  FLOAT_t
ctypedef np.double_t DOUBLE_t
ctypedef np.long_t   LONG_t


#===============================================================================
# Factories for mapping functions for BAMGenomeArray or other structures
# Each factory returns a function that takes a list of pysam.AlignedSegments
# and a GenomicSegment, and returns a list of pysam.AlignedSegments that mapped
# to that interval under the mapping rules specified in that function, and
# a corresponding vector of read counts mapping to each nucleotide position
# in the GenomicSegment
#===============================================================================


cdef class CenterMapFactory:
    """
    CenterMapFactory(nibble=0)

    Read mapping tool for :meth:`BAMGenomeArray.set_mapping`. `nibble`
    positions is removed from each side of each read alignment, and the `N`
    remaining positions are each apportioned `1/N` of the read count, so that
    the entire read is counted once.

    Parameters
    ----------
    nibble : int >= 0, optional
        Number of bases to remove from each side of read (Default: `0`)
    """

    def __cinit__(self, unsigned int nibble = 0):
        """Read mapping tool for :meth:`BAMGenomeArray.set_mapping`.  A
        user-specified number of bases is removed from each side of each read
        alignment, and the `N` remaining bases are each apportioned `1/N` of
        the read count, so that the entire read is counted once.

        Parameters
        ----------
        nibble : int >= 0, optional
            Number of bases to remove from each side of read (Default: `0`)
        """
        if nibble < 0:
            raise ValueError("CenterMapFactory: `nibble` must be >= 0. Got %s." % nibble)

        self.nibble = nibble

    @cython.boundscheck(False) # valid because indices are explicitly checked
    @cython.cdivision(True) # we can do this because we explicitly check map_length > 0
    def __call__(self, list reads not None, GenomicSegment seg not None):
        """Returns reads covering a region, and a count vector mapping reads
        to specific positions in the region. `self.nibble` bases are trimmed from each
        side of the read, and each of the `N` remaining alignment positions are
        incremented by `1/N`

        Parameters
        ----------
        seg : |GenomicSegment|
            Region of interest

        reads : list of :py:class:`pysam.AlignedSegment`
            Reads to map

        Returns
        -------
        list
            List of reads that were mapped into the output array

        :py:class:`numpy.ndarray`
            Vector of counts at each position in `seg`
        """
        cdef bint do_warn = 0

        cdef long seg_start = seg.start
        cdef long seg_end   = seg.end
        cdef long seg_len   = seg_end - seg_start
        cdef unsigned int nibble = self.nibble

        # count array we will return
        cdef np.ndarray count_array = np.zeros(seg_len, dtype=DOUBLE)
        cdef double [:] count_view = count_array

        cdef list reads_out = []
        cdef list read_positions

        cdef:
            AlignedSegment read
            int coord, map_length
            unsigned int read_length, i
            DOUBLE_t val

        for read in reads:
            read_positions = <list>read.positions
            read_length = len(read_positions)
            map_length = read_length - 2 * nibble
            if map_length < 0:
                do_warn = 1
                continue
            elif map_length > 0:
                val = 1.0 / map_length
                for i in range(nibble, read_length - nibble):
                    coord = <long>(read_positions[i]) - seg_start
                    if coord >= 0 and coord < seg_len:
                        count_view[coord] += val

                reads_out.append(read)

        if do_warn == 1:
            warn_onceperfamily(
                "Data contains read alignments shorter than `2*nibble` value of '%s' nt. Ignoring these."
                % (2 * nibble),
                DataWarning
            )

        return reads_out, count_array

    property nibble:
        """Number of positions to trim from each side of read alignment before
        assigning genomic positions.
        """
        def __get__(self):
            return self.nibble

        def __set__(self, unsigned int val):
            self.nibble = val


cdef class FivePrimeMapFactory:
    """
    FivePrimeMapFactory(offset=0)

    Fiveprime mapping factory for :py:meth:`BAMGenomeArray.set_mapping`. Reads
    are mapped at `offset` nucleotides from the fiveprime end of their
    alignment.

    Parameters
    ----------
    offset : int >= 0, optional
        Offset from 5' end of read, in direction of 3' end, at which reads
        should be counted (Default: `0`)
    """

    def __cinit__(self, int offset = 0):
        """Fiveprime mapping factory for :py:meth:`BAMGenomeArray.set_mapping`.
        Reads are mapped at a user-specified offset from the fiveprime end of
        the alignment.

        Parameters
        ----------
        offset : int >= 0, optional
            Offset from 5' end of read, in direction of 3' end, at which
            reads should be counted (Default: `0`)
        """
        if offset < 0:
            raise ValueError("FivePrimeMapFactory: `offset` must be <= 0. Got %s." % offset)
        self.offset = offset

    def __call__(self, list reads not None, GenomicSegment seg not None):
        """Returns reads covering a region, and a count vector mapping reads
        to specific positions in the region, mapping reads at `self.offset`
        from the fiveprime end of each read.

        Parameters
        ----------
        reads : list of :py:class:`pysam.AlignedSegment`
            Reads to map

        seg : |GenomicSegment|
            Region of interest


        Returns
        -------
        list
            List of :py:class:`pysam.AlignedSegment` that map to region

        :py:class:`numpy.ndarray`
            Vector of counts at each position in `seg`
        """
        cdef:
            long seg_start = seg.start
            long seg_end   = seg.end
            long seg_len   = seg_end - seg_start
            np.ndarray count_array = np.zeros(seg_len, dtype=LONG)
            long [:] count_view = count_array

            list reads_out = []
            list read_positions
            AlignedSegment read
            long p_site
            int read_length
            int do_warn = 0
            int read_offset = self.offset

        if seg.c_strand == reverse_strand:
            read_offset = -read_offset - 1

        for read in reads:
            read_positions = <list>read.positions
            read_length = len(read_positions)
            if self.offset >= read_length:
                do_warn = 1
                continue

            p_site = read_positions[read_offset]
            if p_site >= seg_start and p_site < seg_end:
                reads_out.append(read)
                count_view[p_site - seg_start] += 1

        if do_warn == 1:
            warn_onceperfamily(
                "Data contains read alignments shorter than offset (%s nt). Ignoring."
                % (self.offset),
                DataWarning
            )

        return reads_out, count_array

    property offset:
        """Distance from 5' end of read at which to assign reads"""
        def __get__(self):
            return self.offset
        def __set__(self, int val):
            self.offset = val


cdef class ThreePrimeMapFactory:
    """
    ThreePrimeMapFactory(offset=0)

    Threeprime mapping factory for :py:meth:`BAMGenomeArray.set_mapping`.
    Reads are mapped `offset` nucleotides from the threeprime end of their
    alignments.

    Parameters
    ----------
    offset : int >= 0, optional
        Offset from 3' end of read, in direction of 5' end, at which reads
        should be counted (Default: `0`)
    """

    def __cinit__(self, int offset = 0):
        """Threeprime mapping factory for :py:meth:`BAMGenomeArray.set_mapping`.
        Reads are mapped at a user-specified offset from the fiveprime end of
        the alignment.

        Parameters
        ----------
        offset : int >= 0, optional
            Offset from 3' end of read, in direction of 5' end, at which reads
            should be counted (Default: `0`)
        """
        if offset < 0:
            raise ValueError("ThreePrimeMapFactory: `offset` must be <= 0. Got %s." % offset)
        self.offset = offset

    def __call__(self, list reads not None, GenomicSegment seg not None):
        """Returns reads covering a region, and a count vector mapping reads
        to specific positions in the region, mapping reads at `self.offset`
        from the threeprime end of each read.

        Parameters
        ----------
        reads : list of :py:class:`pysam.AlignedSegment`
            Reads to map

        seg : |GenomicSegment|
            Region of interest


        Returns
        -------
        list
            List of :py:class:`pysam.AlignedSegment` that map to region

        :py:class:`numpy.ndarray`
            Vector of counts at each position in `seg`
        """
        cdef:
            long seg_start = seg.start
            long seg_end   = seg.end
            long seg_len   = seg_end - seg_start
            np.ndarray count_array = np.zeros(seg_len, dtype=LONG)
            long [:] count_view = count_array

            list reads_out = []
            list read_positions
            AlignedSegment read
            long p_site
            int read_length
            int do_warn = 0
            int read_offset = self.offset

        if seg.c_strand != reverse_strand:
            read_offset = -read_offset - 1

        for read in reads:
            read_positions = <list>read.positions
            read_length = len(read_positions)
            if self.offset >= read_length:
                do_warn = 1
                continue

            p_site = read_positions[read_offset]
            if p_site >= seg_start and p_site < seg_end:
                reads_out.append(read)
                count_view[p_site - seg_start] += 1

        if do_warn == 1:
            warn_onceperfamily(
                "Data contains read alignments shorter than offset (%s nt). Ignoring."
                % self.offset,
                DataWarning
            )

        return reads_out, count_array


    property offset:
        """Distance from 3' end of read at which to assign reads"""
        def __get__(self):
            return self.offset
        def __set__(self, int val):
            self.offset = val


cdef class VariableFivePrimeMapFactory:
    """
    VariableFivePrimeMapFactory(offset_dict)

    Fiveprime-variable mapping for :py:meth:`BAMGenomeArray.set_mapping`.
    Reads are mapped at a user-specified offsets from the fiveprime end of each
    alignment. The offset for a read of a given length is supplied in `offset_dict[readlen]`

    Parameters
    ----------
    offset_dict : dict
        Dictionary mapping read lengths to offsets that should be applied
        to reads of that length during mapping. A special key, `'default'` may
        be supplied to provide a default value for lengths not specifically
        enumerated in `offset_dict`
    """

    def __cinit__(self, dict offset_dict, *args):
        """
        Parameters
        ----------
        offset_dict : dict
            Dictionary mapping read lengths to offsets that should be applied
            to reads of that length during mapping. A special key, `'default'` may
            be supplied to provide a default value for lengths not specifically
            enumerated in `offset_dict`
        """
        cdef:
            int [:] fw_view = self.forward_offsets
            int [:] rc_view = self.reverse_offsets
            int offset
            object read_length # can be int or str
            int i

        fw_view [:] = _BAD_OFFSET
        rc_view [:] = _BAD_OFFSET

        # this might seem silly, because nobody would every make a
        # FivePrimeVariableMapFactory without an offset_dict, but this allows
        # reuse of this __cinit__ in StratifiedVariableFivePrimeMapFactory.
        if offset_dict is None:
            offset_dict = { "default" : 0 }

        if "default" in offset_dict:
            default = int(offset_dict["default"])
            fw_view[default + 1:] = default
            i = default + 1
            while i < len(rc_view):
                rc_view[i] = i - default - 1
                i += 1

        # store offsets in c arrays for fast lookup
        # value for `default` in 0th place
        for read_length, offset in offset_dict.items():
            if read_length != "default":
                if offset >= read_length:
                    if read_length >= default:
                        warn_onceperfamily("Given offset '%s' longer than read length '%s'. Falling back to default '%s'." % (offset, read_length, default),
                             DataWarning)
                        offset = default
                    else:
                        warn_onceperfamily("Given offset '%s' and default '%s' are longer than read length '%s'. Ignoring %s-mers." % (offset, default, read_length, read_length),
                             DataWarning)
                    continue

                fw_view[read_length] = offset
                rc_view[read_length] = read_length - offset - 1

    @classmethod
    def from_file(cls, object fn_or_fh):
        """
        from_file(fn_or_fh)

        Create a :class:`VariableFivePrimeMapFactory` from a text file.

        The text file should be formatted as by the :mod:`~plastid.bin.psite`
        script: a tab-delimited two-column text file in which the first column
        is a read length (or the special word 'default'), and the second column
        an offset from the 5' end of the read corresponding to that read length.
        For example::

            25      12
            26      12
            27      13
            28      13
            29      14
            30      14
            default 14


        Parameters
        ----------
        fn_or_fh : str or open file-like
            If str, it should be a path to the text file. If file-like, it
            should be an open file-like object pointing to the data that
            would be in the text file.
        """
        if isinstance(fn_or_fh, str):
            with open(fn_or_fh) as my_fh:
                mapper = cls(_parse_variable_offset_file(my_fh))
                my_fh.close()
                return mapper
        else:
            fh = CommentReader(fn_or_fh)

        return cls(_parse_variable_offset_file(fh))

    @cython.boundscheck(False) # valid because indices are explicitly checked in __cinit__
    def __call__(self, list reads not None, GenomicSegment seg not None):
        """Returns reads covering a region, and a count vector mapping reads
        to specific positions in the region, mapping reads at possibly varying
        offsets from the 5' end of each read as determined by `self.offset_dict`

        Parameters
        ----------
        seg : |GenomicSegment|
            Region of interest

        reads : list of :py:class:`pysam.AlignedSegment`
            Reads to map

        Returns
        -------
        list
            List of reads that were mapped into the output array

        :py:class:`numpy.ndarray`
            Vector of counts at each position in `seg`
        """
        cdef:
            int [:] offsets = self.forward_offsets

            long seg_start = seg.start
            long seg_end   = seg.end
            long seg_len   = seg_end - seg_start

            int no_offset_length
            int do_no_offset_warning = 0

            int            read_length, offset
            long           p_site
            AlignedSegment read

            np.ndarray count_array = np.zeros(seg_len, dtype=LONG)
            long [:]   count_view  = count_array
            list read_positions
            list reads_out       = []

        if seg.c_strand == reverse_strand:
            offsets = self.reverse_offsets

        for read in reads:
            read_positions = <list>read.positions
            read_length    = len(read_positions)
            offset         = offsets[read_length]

            if offset == _BAD_OFFSET:
                do_no_offset_warning = 1
                no_offset_length = read_length
                continue

            p_site = read_positions[offset]
            if p_site >= seg_start and p_site < seg_end:
                reads_out.append(read)
                count_view[p_site - seg.start] += 1

        if do_no_offset_warning == 1:
            warn_onceperfamily(
                "No usable offset for reads of length %s nt in offset dict. Ignoring these."
                % (no_offset_length),
                DataWarning
            )

        return reads_out, count_array


cdef class StratifiedVariableFivePrimeMapFactory(VariableFivePrimeMapFactory):
    """
    StratifiedVariableFivePrimeMapFactory(offset_dict, min = 25, max = 35)

    Fiveprime-variable mapping for :py:meth:`BAMGenomeArray.set_mapping`, stratified by read length.

    Reads are mapped into a 2D array of counts at each position in a region of
    interest (column), stratified by read length (row). Mapping is performed
    by applying a user-specified offset from the fiveprime end of each alignment.
    A unique offset may be supplied for each read length.

    Reads outside of pre-specified minimum and maximum lengths are ignored.


    Parameters
    ----------
    offset_dict : dict or `None`
        Dictionary mapping read lengths to offsets that should be applied
        to reads of that length during mapping. A special key, `'default'` may
        be supplied to provide a default value for lengths not specifically
        enumerated in `offset_dict.` If `None`, all reads are mapped to their
        5' ends.

    min_length : int
        Minimum length read to map

    max_length : int
        Maximum length read to map, inclusive


    Attributes
    ----------
    row_keys : numpy.ndarray
        1D numpy array indicating which read length is assigned to which row

    shape : list of integers
        size of each axis of output, excluding the final axis, which is
        specified the length of incoming GenomicSegments.

     .. warning::

        The :meth:`~plastid.genomics.genome_array.BAMGenomeArray.to_variable_step`,
        :meth:`~plastid.genomics.genome_array.BAMGenomeArray.to_bedgraph`,
        :meth:`~plastid.genomics.genome_array.BAMGenomeArray.to_genome_array`,
        of :class:`BAMGenoemArray` do not support multi-dimensional map
        factories.-
    """
    def __cinit__(self, dict offset_dict, int min = 25, int max = 35):
        """
        Parameters
        ----------
        offset_dict : dict
            Dictionary mapping read lengths to offsets that should be applied
            to reads of that length during mapping. A special key, `'default'` may
            be supplied to provide a default value for lengths not specifically
            enumerated in `offset_dict`

        min_length : int
            Minimum length read to map

        max_length : int
            Maximum length read to map, inclusive
        """
        if max <= min:
            raise ValueError("Max length '%s' must be >= min length '%s'. " % (max, min))

        self.min_length  = min
        self.max_length  = max
        self._numlengths = max - min + 1

    @cython.boundscheck(False) # valid because indices are explicitly checked in VariableFivePrimeMapFactory.__cinit__
    def __call__(self, list reads not None, GenomicSegment seg not None):
        """
        Map reads covering `seg` into a 2D count array, in which reads at each
        nucleotide position (column) are stratified by read length (row), and,
        optionally, offset from their 5' ends.

        Parameters
        ----------
        seg : |GenomicSegment|
            Region of interest

        reads : list of :py:class:`pysam.AlignedSegment`
            Reads to map

        Returns
        -------
        list
            List of reads that were mapped into the output array

        :py:class:`numpy.ndarray`
            2D array, in which each cell value is the number of counts of a
            given length (row) at  given position (column)
        """
        cdef:
            long seg_start             = seg.start
            long seg_end               = seg.end
            long seg_len               = seg_end - seg_start

            int [10000] offsets        = self.forward_offsets
            int num_lengths            = self._numlengths
            int min_length             = self.min_length

            list reads_out             = []
            int read_length, offset, p_site

            AlignedSegment read

            np.ndarray count_array = np.zeros((num_lengths, seg_len), dtype=LONG)
            long [:, :] count_view  = count_array
            long psite

        if seg.c_strand == reverse_strand:
            offsets = self.reverse_offsets

        for read in reads:
            read_positions = <list>read.positions
            read_length = len(read_positions)
            if read_length >= min_length and read_length <= self.max_length:

                offset = offsets[read_length]
                p_site = read_positions[offset]

                if p_site >= seg_start and p_site < seg_end:
                    reads_out.append(read)
                    count_view[read_length - min_length, p_site - seg_start] += 1

        return reads_out, count_array

    property row_keys:
        """numpy array of read lengths corresponding to each row of mapped data."""
        def __get__(self):
            return np.arange(self.min_length, self.max_length+1)

    property shape:
        """list containing size of each axis of output, excluding the final axis,
        which is specified the length of incoming GenomicSegments."""
        def __get__(self):
            return [self._numlengths]


cdef class SizeFilterFactory:
    """
    SizeFilterFactory(min = 1, max = -1)

    Create a read-length filter can be applied at runtime to a |BAMGenomeArray|
    using ::meth:`BAMGenomeArray.add_filter`.

    Parameters
    ----------
    min : float, optional
        Minimum read length to pass filter, inclusive (Default: `1`)

    max : float or numpy.inf, optional
        Maximum read length to pass filter, inclusive. If set to `-1`, then
        there is no maximum length filter. (Default: -1, no filter)
    """

    def __cinit__(self, int min = 1, int max = -1):
        """Create a read-length filter can be applied at runtime to a |BAMGenomeArray|
        using ::meth:`BAMGenomeArray.add_filter`

        Parameters
        ----------
        min : int >= 1, optional
            Minimum read length to pass filter, inclusive (Default: `1`)

        max : int, optional
            Maximum read length to pass filter, inclusive. If `-1`, then
            there is no maximum length filter. (Default: `-1`, no filter.)

        Returns
        -------
        function
        """
        if max != -1 and max < min:
            raise ValueError("Alignment size filter: max read length must be >= min read length")

        if min < 1:
            raise ValueError("Alignment size filter: min read length must be >= 1. Got %s" % min)

        self.min_ = min
        self.max_ = max

    def __call__(self, AlignedSegment read not None):
        cdef unsigned int my_length = len(read.positions)
        return my_length >= self.min_ and (my_length <= self.max_ or self.max_ == -1)

